# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Structured data store type."""

from typing import Any, Dict, Optional, Tuple, Union

from ..data import Data
from ._member import Member
from ._meta import StructureMeta
from ._structureview import StructureView


class Structure(list, Data, metaclass=StructureMeta):

    """Structured data store type."""

    # filled in by metaclass
    __ignore_flags__: Tuple[bool, ...] = ()
    __members__: Tuple[Member, ...] = ()
    __nbytes__: Union[None, int] = 0
    __offsets__: Union[None, Tuple[int, ...]] = ()
    __byteorder__: str = "little"
    __implementation__: Optional[str] = None

    # Metaclass generates the following methods for each subclass:
    #   __init__
    #   member getter/setters
    #   __eq__
    #   __ne__
    #   __pack__
    #   __pack_and_dump__
    #   __repr__
    #   __unpack__
    #   __unpack_and_dump__

    def asdict(self) -> Dict[str, object]:
        """Return structure members in dictionary form.

        :returns: structure members
        :rtype: dict

        """
        return {
            name: getattr(self, name)
            for name in (member.name for member in self.__members__)
        }

    @classmethod
    def _make_structure_from_dict(cls, members: Dict[str, Any]) -> "Structure":
        return cls(**members)

    def __setattr__(self, name, value):
        # get the attribute to raise an exception if invalid name
        getattr(self, name)
        object.__setattr__(self, name, value)

    @classmethod
    def __view__(cls, buffer, offset=0):
        """Create plum view of bytes buffer.

        :param buffer: bytes buffer
        :type buffer: bytes-like (e.g. bytes, bytearray, memoryview)
        :param int offset: byte offset

        """
        if not cls.__nbytes__:
            raise TypeError(
                f"cannot create view for structure {cls.__name__!r} "
                "with variable size"
            )

        return StructureView(cls, buffer, offset)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test base view type exceptions."""

import pytest

from baseline import Baseline

from plum.bigendian import uint16
from plum.conformance import wrap_message
from plum.exceptions import UnpackError
from plum.view import PlumView


class TestExceptions:

    """Test base view type exceptions."""

    def test_init_bad_plumtype(self):
        """Verify initializing without a valid plum type raises an exception."""
        with pytest.raises(TypeError) as trap:
            PlumView(fmt=0, buffer=bytearray(), offset=0)

        expected = Baseline("""invalid plumtype""")

        assert wrap_message(trap.value) == expected

    def test_pack_issues(self):
        """Verify initializing without a valid plum type raises an exception."""
        expected = Baseline(
            """
            +--------+----------------------+-------+--------+
            | Offset | Value                | Bytes | Format |
            +--------+----------------------+-------+--------+
            | 0      | <insufficient bytes> | 00    | uint16 |
            +--------+----------------------+-------+--------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack 'uint16' (2 needed, only 1 available)
            """
        )

        view = uint16.view(b"\x00")

        with pytest.raises(UnpackError) as trap:
            view.pack()

        assert wrap_message(trap.value) == expected

    def test_hash(self):
        expected = Baseline(
            """
            unhashable type: IntView
            """
        )

        view = uint16.view(b"\x00\x00")

        with pytest.raises(TypeError) as trap:
            hash(view)

        assert wrap_message(trap.value) == expected



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test integer enumeration transform properties."""

from plum.flag import FlagX

from sample_flag import Register


class TestDefault:

    """Test with as many left to default as possible."""

    register = FlagX("register", Register, 1)

    def test_name(self):
        assert self.register.name == "register"

    def test_enum(self):
        assert self.register.enum is Register

    def test_nbytes(self):
        assert self.register.nbytes == 1

    def test_byteorder(self):
        assert self.register.byteorder == "little"


class TestPositional:

    """Test explicitly defined with positional argument."""

    register = FlagX("register", Register, 2, "big")

    def test_name(self):
        assert self.register.name == "register"

    def test_enum(self):
        assert self.register.enum is Register

    def test_nbytes(self):
        assert self.register.nbytes == 2

    def test_byteorder(self):
        assert self.register.byteorder == "big"


class TestKeyword(TestPositional):

    """Test explicitly defined with keyword argument."""

    register = FlagX(name="register", enum=Register, nbytes=2, byteorder="big")



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test items transform properties."""

import pytest

from plum.exceptions import SizeError
from plum.items import ItemsX
from plum.littleendian import uint8


class TestDefault:

    """Test with as many left to default as possible."""

    itemsx = ItemsX("itemsx")

    def test_name(self):
        assert self.itemsx.name == "itemsx"

    def test_nbytes(self):
        with pytest.raises(SizeError):
            return self.itemsx.nbytes

    def test_fmt(self):
        assert self.itemsx.fmt is None


class TestPositional:

    """Test explicitly defined with positional argument."""

    itemsx = ItemsX("itemsx", uint8)

    def test_name(self):
        assert self.itemsx.name == "itemsx"

    def test_nbytes(self):
        assert self.itemsx.nbytes == 1

    def test_fmt(self):
        assert self.itemsx.fmt is uint8


class TestKeyword(TestPositional):

    """Test explicitly defined with keyword argument."""

    itemsx = ItemsX(name="itemsx", fmt=uint8)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""BitFields __init__ and property methods generator."""

from enum import EnumMeta
from typing import Any, Dict, Generator, Tuple, ValuesView

from .._code_injector import CodeInjector
from ..enum import EnumX
from ._bitfield import BitField

INDENT = "    "


class CodeMaker:

    """BitFields method code generator."""

    bitfields: Tuple[BitField, ...]

    def __init__(
        self,
        bitfields: ValuesView[BitField],
        code_injector: CodeInjector,
    ) -> None:
        self.bitfields = tuple(bitfields)

        type_names = [
            code_injector.get_expression(bitfield.typ, f"{bitfield.name}.typ")
            for bitfield in bitfields
        ]

        self.arg_type_hints = {}
        for bitfield, type_name in zip(bitfields, type_names):
            if bitfield.typ is int:
                self.arg_type_hints[bitfield.name] = "int"
            else:
                self.arg_type_hints[bitfield.name] = f"Union[int, {type_name}]"

        self.getter_type_hints = {}
        for bitfield, expression in zip(bitfields, type_names):
            if isinstance(bitfield.typ, EnumX) and not bitfield.typ.strict:
                expression = f"Union[int, {expression}]"
            self.getter_type_hints[bitfield.name] = expression

        self.type_expressions = {}
        for bitfield, expression in zip(bitfields, type_names):
            if "." in expression:
                expression = f"type(self).{expression}"

            self.type_expressions[bitfield.name] = expression

    def iter_init_lines(self, default: int) -> Generator[str, None, None]:
        """Generate __init__ method lines."""
        parameters = []

        for bitfield in self.bitfields:

            parameter = f"{bitfield.name}: {self.arg_type_hints[bitfield.name]}"

            if bitfield.default is not None:
                parameter += f" = {bitfield.default}"

            parameters.append(parameter)

        args = ", ".join(parameters)

        if args:
            yield f"def __init__(self, *, {args}) -> None:"
        else:
            yield "def __init__(self) -> None:"
        yield INDENT + f"self.__value__ = {default}"

        for bitfield in self.bitfields:
            yield INDENT + f"self.{bitfield.name} = {bitfield.name}"

        yield ""

    def iter_getter_lines(self, bitfield: BitField):
        """Generate getter."""
        yield f"@{bitfield.name}.getter"
        yield f"def {bitfield.name}(self) -> {self.getter_type_hints[bitfield.name]}:"
        if bitfield.doc:
            yield INDENT + f'"""{bitfield.doc}"""'

        if bitfield.lsb:
            retval = f"(int(self) >> {bitfield.lsb}) & {bitfield.mask}"
        else:
            retval = f"int(self) & {bitfield.mask}"

        if bitfield.signed:
            yield INDENT + f"value = {retval}"
            yield INDENT + f"value = -((1 << {bitfield.size}) - value) if {bitfield.signbit} & value else value"
            retval = "value"

        if hasattr(bitfield.typ, "__fields__"):
            yield INDENT + f"value = {self.type_expressions[bitfield.name]}.from_int({retval})"
            yield INDENT + "try:"
            yield INDENT + "    bitoffset, store = self.__bitoffset_store__"
            yield INDENT + "except AttributeError:"
            yield INDENT + f"    bitoffset, store = {bitfield.lsb}, self"
            if bitfield.lsb:
                yield INDENT + "else:"
                yield INDENT + f"    bitoffset += {bitfield.lsb}"
            yield INDENT + "value.__bitoffset_store__ = bitoffset, store"
            retval = "value"

        elif bitfield.typ is not int:
            retval = f"{self.type_expressions[bitfield.name]}({retval})"

        yield INDENT + f"return {retval}"
        yield ""

    def iter_setter_lines(self, bitfield: BitField, nested: bool):
        """Generate setter."""
        typ = bitfield.typ
        typ_name = self.type_expressions[bitfield.name]

        yield f"@{bitfield.name}.setter"
        yield f"def {bitfield.name}(self, value: {self.arg_type_hints[bitfield.name]}) -> None:"
        if bitfield.doc:
            yield INDENT + f'"""{bitfield.doc}"""'

        if hasattr(bitfield.typ, "__fields__"):
            yield INDENT + f"value = int({typ_name}.from_int(value))"
        elif bitfield.typ is int:
            yield INDENT + "value = int(value)"
        else:
            yield INDENT + f"value = int({typ_name}(value))"

        if (
            not hasattr(typ, "__fields__")
            and not isinstance(typ, EnumMeta)
            and typ is not bool
        ):
            yield INDENT + f"if not ({bitfield.minvalue} <= value <= {bitfield.maxvalue}):"
            yield INDENT + f'    raise ValueError("bit field {bitfield.name!r} requires {bitfield.minvalue} <= number <= {bitfield.maxvalue}")'

        if nested:
            yield INDENT + "try:"
            yield INDENT + "    bitoffset, store = self.__bitoffset_store__"
            yield INDENT + "except AttributeError:"
            yield INDENT + f"    bitoffset, store = {bitfield.lsb}, self"
            if bitfield.lsb:
                yield INDENT + "else:"
                yield INDENT + f"    bitoffset += {bitfield.lsb}"
            yield INDENT + f"store.__value__ = (store.__value__ & ~({bitfield.mask} << bitoffset)) | ((value & {bitfield.mask}) << bitoffset)"
        elif bitfield.lsb:
            yield INDENT + f"self.__value__ = (self.__value__ & {~(bitfield.mask << bitfield.lsb)}) | ((value & {bitfield.mask}) << {bitfield.lsb})"
        else:
            yield INDENT + f"self.__value__ = (self.__value__ & {~bitfield.mask}) | (value & {bitfield.mask})"
        yield ""

    def iter_repr_lines(self) -> Generator[str, None, None]:
        """Bitfields __repr__ method generator."""
        arg_reprs = ", ".join(f.argrepr for f in self.bitfields if f.argrepr)
        class_name = "{type(self).__name__}"

        yield "def __repr__(self) -> str:"
        yield "    try:"
        yield f'        return f"{class_name}({arg_reprs})"'
        yield "    except Exception:"
        yield f'        return f"{class_name}()"'
        yield ""

    def iter_lines(
        self, namespace: Dict[str, Any], default: int, nested: bool
    ) -> Generator[str, None, None]:
        """Generate BitFields class method code lines."""
        if "__init__" not in namespace:
            yield from self.iter_init_lines(default)

        for bitfield in self.bitfields:
            if bitfield.fget is None:
                yield from self.iter_getter_lines(bitfield)
            if bitfield.fset is None and not bitfield.readonly:
                yield from self.iter_setter_lines(bitfield, nested)

        if "__repr__" not in namespace:
            yield from self.iter_repr_lines()



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Integer flag to bytes and bytes to integer flag transform."""

from enum import IntFlag
from typing import List, Optional, Tuple, Type

from ._getbytes import getbytes
from ._typing import ByteOrderHint
from .dump import Record
from .exceptions import InsufficientMemoryError
from .transform import Transform


class FlagX(Transform):

    """Flag Enumeration to bytes and bytes to flag enumeration transform."""

    __ben__: Tuple[ByteOrderHint, Type[IntFlag], int]

    def __init__(
        self,
        name: str,
        enum: Type[IntFlag],
        nbytes: int,
        byteorder: ByteOrderHint = "little",
    ) -> None:
        super().__init__(name)

        assert nbytes > 0
        assert byteorder in {"big", "little"}

        self.__nbytes__ = nbytes
        self.__ben__ = byteorder, enum, nbytes

    @property
    def byteorder(self) -> ByteOrderHint:
        """Byte order ("little" or "big")."""
        byteorder, _enum, _nbytes = self.__ben__
        return byteorder

    @property
    def enum(self) -> Type[IntFlag]:
        """Integer flag enumeration."""
        _byteorder, enum, _nbytes = self.__ben__
        return enum

    def __pack__(
        self, value: int, pieces: List[bytes], dump: Optional[Record] = None
    ) -> None:
        byteorder, _enum, nbytes = self.__ben__

        if dump is None:
            pieces.append(int.to_bytes(value, nbytes, byteorder, signed=False))
        else:
            try:
                int_value = int(value)
                piece = int.to_bytes(int_value, nbytes, byteorder, signed=False)

                dump.memory = piece

                self._add_flags_to_dump(int_value, dump)

            except Exception:
                # use repr in case str or something that otherwise looks like an int
                dump.value = repr(value)
                raise

            pieces.append(piece)

    def __unpack__(
        self, buffer: bytes, offset: int, dump: Optional[Record] = None
    ) -> Tuple[IntFlag, int]:
        byteorder, enum, nbytes = self.__ben__

        if dump is None:
            end = offset + nbytes

            if len(buffer) < end:
                raise InsufficientMemoryError("too few bytes to unpack")

            return (
                enum(int.from_bytes(buffer[offset:end], byteorder, signed=False)),
                end,
            )

        piece, offset = getbytes(buffer, offset, dump, nbytes)

        int_value = int.from_bytes(piece, byteorder, signed=False)

        value = enum(int_value)

        dump.value = value

        self._add_flags_to_dump(int_value, dump)

        return value, offset

    def _add_flags_to_dump(self, value: int, dump: Record) -> None:
        _byteorder, enum, _nbytes = self.__ben__

        dump.value = value

        for member in enum:
            member_name = member.name or "unlikely"  # mypy thinks it might be None
            dump.add_record(
                access="." + member_name.lower(),
                bits=(int(member.bit_length() - 1), 1),
                value=bool(value & member),
                fmt="bool",
            )



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test construction of bit field classes."""

# pylint: disable=invalid-name

import pytest
from baseline import Baseline

from plum.structure import Structure, bitfield_member
from plum.conformance import wrap_message


class TestExceptions:

    """Test validation of class."""

    @staticmethod
    def test_bitfield_bad_type():
        """Test annotation contains bad type."""

        expected_message = Baseline(
            """
            bit field type must be int-like
            """
        )

        with pytest.raises(TypeError) as trap:

            class Sample(Structure):  # pylint: disable=unused-variable

                f1: int = bitfield_member(lsb=0, size=8)
                f2 = bitfield_member(lsb=8, size=8, typ=str)

        assert wrap_message(trap.value) == expected_message

    @staticmethod
    def test_bitfield_overlap():
        """Test bit fields overlap."""

        expected_message = Baseline(
            """
            bit field 'f2' overlaps with 'f1'
            """
        )

        with pytest.raises(TypeError) as trap:

            class Sample(Structure):  # pylint: disable=unused-variable

                f1: int = bitfield_member(lsb=0, size=8)
                f2: int = bitfield_member(lsb=7, size=8)

        assert wrap_message(trap.value) == expected_message

    @staticmethod
    def test_nbytes_too_small():
        """Test nbytes argument insufficient size."""

        expected_message = Baseline(
            """
            'nbytes' must be at least 2 for bitfield 'f1' (to have sufficient room
            for it and those that follow)
            """
        )

        with pytest.raises(TypeError) as trap:

            class Sample(Structure):  # pylint: disable=unused-variable

                f1: int = bitfield_member(lsb=0, size=8, nbytes=1)
                f2: int = bitfield_member(lsb=8, size=8)

        assert wrap_message(trap.value) == expected_message



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Structure of uniquely named and typed members data store and transform."""

from ._bitfield_member import bitfield_member
from ._dimmed_member import dimmed_member
from ._member import member
from ._meta import StructureMeta
from ._sized_member import sized_member
from ._structure import Structure

bitfield_member.__module__ = "plum.structure"
dimmed_member.__module__ = "plum.structure"
member.__module__ = "plum.structure"
StructureMeta.__module__ = "plum.structure"
sized_member.__module__ = "plum.structure"
Structure.__module__ = "plum.structure"



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Sample integer enumeration"""

from enum import IntEnum


class Register(IntEnum):

    """Tested class."""

    PC = 0
    SP = 1
    R0 = 2
    R1 = 3



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Bit fields data store type."""

from typing import Dict, Optional
from .._typing import ByteOrderHint
from ..data import DataMeta
from .._code_injector import CodeInjector
from ._bitfield import BitField
from ._code_maker import CodeMaker


class BlendedDefault:

    """Fake BitFields type for calculating default value."""

    value: int

    def __init__(self, value: int) -> None:
        self.value = value

    def add(self, bitfield: BitField) -> None:
        """Add bitfield default value."""
        if bitfield.default is not None:
            mask = bitfield.mask
            lsb = bitfield.lsb
            self.value = (self.value & ~(mask << lsb)) | (
                (bitfield.default & mask) << lsb
            )


class BitFieldsMeta(DataMeta):

    """Bit fields data store type metaclass."""

    __nested__: bool
    __ignore__: int
    __byteorder__: ByteOrderHint
    __default__: bool

    def __new__(
        mcs,
        name,
        bases,
        namespace,
        nbytes: Optional[int] = None,
        byteorder: ByteOrderHint = "little",
        default: int = 0,
        ignore: int = 0,
        nested: bool = False,
        fieldorder: str = "most_to_least",
    ):
        # pylint: disable=too-many-arguments,too-many-locals
        # pylint: disable=too-many-branches,too-many-statements
        if byteorder not in {"big", "little"}:
            raise ValueError('byteorder must be either "big" or "little"')

        if fieldorder not in {"least_to_most", "most_to_least"}:
            raise ValueError(
                'fieldorder must be either "least_to_most" or "most_to_least"'
            )

        # validate bit field class attributes

        fields = {}
        for fieldname, bitfield in namespace.items():
            if not isinstance(bitfield, BitField):
                continue

            bitfield.finish_initialization(fieldname)

            fields[fieldname] = bitfield

        # fill in unassigned 'pos' in all bit field properties

        lsb = 0
        field_properties = (
            fields.values()
            if fieldorder == "least_to_most"
            else reversed(list(fields.values()))
        )
        for field in field_properties:
            if field._lsb is None:
                field._lsb = lsb
            lsb = field.lsb + field.size

        # check for overlap

        claimed_bits: Dict[int, str] = {}
        for fieldname, field in fields.items():
            for i in range(field.lsb, field.lsb + field.size):
                if i in claimed_bits:
                    raise TypeError(
                        f"bit field {fieldname!r} overlaps with {claimed_bits[i]!r}"
                    )

                claimed_bits[i] = fieldname

        # default/validate 'nbytes'

        if fields:
            numbits = max(field.lsb + field.size for field in fields.values())
            nbytes_needed = (numbits + 7) // 8
        else:
            nbytes_needed = 1

        if nbytes is None:
            nbytes = nbytes_needed
        else:
            nbytes = int(nbytes)
            if nbytes < nbytes_needed:
                raise TypeError(
                    f"nbytes must be at least {nbytes_needed} for bitfields specified"
                )

        max_ = (1 << (nbytes * 8)) - 1

        # validate 'default' and blend in defaults from individual fields

        default = int(default)

        if not 0 <= default <= max_:
            raise ValueError(f"default must be: 0 <= number <= 0x{max_:x}")

        blended_default = BlendedDefault(default)
        for field in fields.values():
            blended_default.add(field)

        # validate 'ignore' and blend in ignores from individual fields

        ignore = int(ignore)

        if not 0 <= ignore <= max_:
            raise ValueError(f"ignore must be: 0 <= number <= 0x{max_:x}")

        for field in fields.values():
            if field.ignore:
                ignore |= field.mask << field.lsb
            elif isinstance(field.typ, BitFieldsMeta):
                ignore |= field.typ.__ignore__ << field.lsb

        namespace["__byteorder__"] = byteorder
        namespace["__compare_mask__"] = (max_ ^ ignore) & max_
        namespace["__fields__"] = fields
        namespace["__default__"] = blended_default.value
        namespace["__fields_with_defaults__"] = {
            name for name, field in fields.items() if field.default is not None
        }
        namespace["__ignore__"] = ignore
        namespace["__max__"] = max_
        namespace["__nbytes__"] = nbytes
        namespace["__nested__"] = bool(nested)

        code_injector = CodeInjector(namespace)

        code_maker = CodeMaker(fields.values(), code_injector)
        lines = list(code_maker.iter_lines(namespace, blended_default.value, nested))

        code_injector.update_script(lines)
        code_injector.execute_lines(lines)
        namespace["__implementation__"] = "\n".join(lines)

        # pylint: disable=too-many-arguments,unused-argument
        return super().__new__(mcs, name, bases, namespace)

    @property
    def byteorder(cls) -> ByteOrderHint:
        """Byte order ("little" or "big")."""
        return cls.__byteorder__

    @property
    def default(cls) -> int:
        """Default integer basis."""
        return cls.__default__

    @property
    def ignore(cls) -> int:
        """Mask for bits to ignore (e.g. 2 ignores bit 1)."""
        return cls.__ignore__

    @property
    def nested(cls) -> bool:
        """Supports being nested in another bitfields."""
        return cls.__nested__



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test OptionalX transform properties."""

import pytest
from plum.exceptions import SizeError
from plum.littleendian import uint8
from plum.optional import OptionalX
from plum.structure import Structure, member


class Sample(Structure):

    member1: int = member(fmt=uint8)


class Test:

    optional_x = OptionalX(fmt=uint8)

    def test_default_name_transform(self):
        assert self.optional_x.name == "Optional[uint8]"

    def test_default_name_structure(self):
        assert OptionalX(fmt=Sample).name == "Optional[Sample]"

    def test_fmt(self):
        assert self.optional_x.fmt is uint8

    def test_explicit_name(self):
        assert OptionalX(fmt=uint8, name="Optional").name == "Optional"

    def test_nbytes(self):
        with pytest.raises(SizeError):
            self.optional_x.nbytes  # pylint: disable=pointless-statement



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test StrX transform properties."""

import pytest

from plum.exceptions import SizeError
from plum.str import StrX


class TestDefault:

    """Test with as many left to default as possible."""

    strx = StrX("strx", "ascii")

    def test_name(self):
        assert self.strx.name == "strx"

    def test_encoding(self):
        assert self.strx.encoding == "ascii"

    def test_errors(self):
        assert self.strx.errors == "strict"

    def test_nbytes(self):
        with pytest.raises(SizeError):
            return self.strx.nbytes

    def test_pad(self):
        assert self.strx.pad == b""

    def test_zero_termination(self):
        assert self.strx.zero_termination is False


class TestPositional:

    """Test explicitly defined with positional argument."""

    strx = StrX("strx", "ascii", "ignore", 10, b"\x00", True)

    def test_name(self):
        assert self.strx.name == "strx"

    def test_encoding(self):
        assert self.strx.encoding == "ascii"

    def test_errors(self):
        assert self.strx.errors == "ignore"

    def test_nbytes(self):
        assert self.strx.nbytes == 10

    def test_pad(self):
        assert self.strx.pad == b"\x00"

    def test_zero_termination(self):
        assert self.strx.zero_termination is True


class TestKeyword(TestPositional):

    """Test explicitly defined with keyword argument."""

    strx = StrX(
        name="strx",
        encoding="ascii",
        errors="ignore",
        nbytes=10,
        pad=b"\x00",
        zero_termination=True,
    )



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test operations for structure view to bytes buffer."""

from collections.abc import Sequence

import pytest
from baseline import Baseline

from plum.conformance import wrap_message
from plum.array import ArrayX
from plum.int import IntView
from plum.littleendian import uint8, uint16
from plum.structure import Structure, member
from plum.structure._structureview import StructureView
from plum.utilities import pack


class FourElementStruct(Structure):

    """Sample structure with four members."""

    m1: int = member(fmt=uint8)
    m2: int = member(fmt=uint8)
    m3: int = member(fmt=uint8)
    m4: int = member(fmt=uint8)


class MyStruct(Structure):

    """Sample structure with two members."""

    m1: int = member(fmt=uint8)
    m2: int = member(fmt=uint16)


class StructWithIgnore(Structure):

    """Sample structure with member that is ignored."""

    not_ignored: int = member(fmt=uint8, default=1)
    ignored: int = member(fmt=uint8, default=0, ignore=True)


class VariableSizeStruct(Structure):

    """Sample variably sized Structure."""

    m1: list = member(fmt=ArrayX("array", fmt=uint8))


class TestExceptions:

    """Test exception corner cases."""

    def test_variable_sized(self):
        """Test view of structure with variable size not allowed."""
        with pytest.raises(TypeError) as trap:
            VariableSizeStruct.view(bytearray(5), offset=1)

        expected = Baseline(
            """
            cannot create view for structure 'VariableSizeStruct' with variable
            size
            """
        )

        assert wrap_message(trap.value) == expected


class TestStructureView:

    """Test operations for structure view to bytes buffer."""

    def test_attribute_access(self):
        """Test accessing members with attribute syntax."""
        memory_backing = bytearray([0xFF, 0x02, 0x05, 0x00, 0xFF])
        struct_view = MyStruct.view(memory_backing, offset=1)

        assert struct_view.m1 == 2
        assert struct_view.m2 == 5

        assert isinstance(struct_view.m1, IntView)
        assert isinstance(struct_view.m2, IntView)

        struct_view.m1 = 1
        assert isinstance(struct_view.m1, IntView)  # verify view not overwritten
        assert struct_view == MyStruct(m1=1, m2=5)
        assert memory_backing == bytearray(b"\xff\x01\x05\x00\xff")

        struct_view.m2 = 3
        assert isinstance(struct_view.m2, IntView)  # verify view not overwritten
        assert struct_view == MyStruct(m1=1, m2=3)
        assert memory_backing == bytearray(b"\xff\x01\x03\x00\xff")

    def test_item_access(self):
        """Test accessing members with item/index syntax."""
        memory_backing = bytearray([0x03, 0x02, 0x01, 0xFF])
        struct_view = MyStruct.view(memory_backing, offset=0)

        assert struct_view[0] == 3
        assert struct_view[1] == 258

        assert isinstance(struct_view[0], IntView)
        assert isinstance(struct_view[1], IntView)

        struct_view[0] = 15
        assert isinstance(struct_view[0], IntView)  # verify view not overwritten
        assert struct_view == MyStruct(m1=15, m2=258)
        assert memory_backing == bytearray(b"\x0f\x02\x01\xff")

        struct_view[1] = 0
        assert isinstance(struct_view[1], IntView)  # verify view not overwritten
        assert struct_view == MyStruct(m1=15, m2=0)
        assert memory_backing == bytearray(b"\x0f\x00\x00\xff")

    def test_representations(self):
        """Test repr and dump representations."""
        memory_backing = bytearray([0xFF, 0x01, 0x02, 0x00, 0xFF])
        struct_view = MyStruct.view(memory_backing, offset=1)
        expected = Baseline(
            """
        +--------+--------+-------+-------+----------+
        | Offset | Access | Value | Bytes | Format   |
        +--------+--------+-------+-------+----------+
        |        |        |       |       | MyStruct |
        | 1      | m1     | 1     | 01    | uint8    |
        | 2      | m2     | 2     | 02 00 | uint16   |
        +--------+--------+-------+-------+----------+
        """
        )

        assert repr(struct_view) == "<view at 0x1: MyStruct(m1=1, m2=2)>"
        assert struct_view.asdict() == {"m1": 1, "m2": 2}
        assert str(struct_view.dump) == expected

    def test_comparison_behavior(self):
        """Test comparison behavior."""
        # pylint: disable=unneeded-not
        memory_backing = bytearray([0xFF, 0x01, 0x02, 0x00, 0xFF])
        my_struct_view = MyStruct.view(memory_backing, offset=1)
        struct_with_ignore_view = StructWithIgnore.view(memory_backing, offset=1)

        assert my_struct_view == [0x01, 0x02]  # list compare format
        assert my_struct_view.copy() == [1, 2]  # list from copy method
        assert struct_with_ignore_view == StructWithIgnore(
            not_ignored=1
        )  # missing member with ignore
        assert (
            struct_with_ignore_view == StructWithIgnore()
        )  # missing member with defaulting

        assert my_struct_view >= [1, 2]
        assert my_struct_view >= [1, 1]
        assert my_struct_view > [1, 1]
        assert not my_struct_view > [2, 2]

        assert my_struct_view <= [1, 2]
        assert my_struct_view <= [2, 2]
        assert my_struct_view < [2, 2]
        assert not my_struct_view < [0, 0]

        assert len(my_struct_view) == 2

    def test_arithmetic_behavior(self):
        """Test arithmetic behavior."""
        memory_backing = bytearray([0xFF, 0x01, 0x02, 0x00, 0xFF])
        my_struct_view = MyStruct.view(memory_backing, offset=1)

        assert my_struct_view + [3, 4] == [1, 2, 3, 4]
        assert [0] + my_struct_view == [0, 1, 2]

        assert my_struct_view * 2 == [1, 2, 1, 2]
        assert 3 * my_struct_view == [1, 2, 1, 2, 1, 2]

    def test_sort_behavior(self):
        """Test sort method behavior."""
        buffer = bytearray([0xFF, 0x0A, 0x0C, 0x03, 0x06, 0xFF])
        struct_view = FourElementStruct.view(buffer, offset=1)

        struct_view.sort()

        assert struct_view == [3, 6, 10, 12]
        assert struct_view.index(3) == 0
        assert struct_view.index(6) == 1
        assert struct_view.index(10) == 2
        assert struct_view.index(12) == 3
        assert buffer == bytearray(b"\xff\x03\x06\x0a\x0c\xff")

    def test_count_behavior(self):
        """Test count method behavior."""
        buffer = bytearray([0xFF, 0x01, 0x01, 0x02, 0x01, 0xFF])
        struct_view = FourElementStruct.view(buffer, offset=1)

        assert struct_view.count(1) == 3
        assert struct_view.count(2) == 1
        assert struct_view.count(3) == 0

    def test_inheritance_behavior(self):
        """Verify ``isinstance`` behavior."""
        struct_view = MyStruct.view(bytearray([0x00, 0x00, 0x00, 0x00]))

        assert isinstance(struct_view, StructureView)
        assert isinstance(struct_view, Sequence)

    def test_pack_pos_argument(self):
        """Test pack() utility supports view as positional argument."""
        buffer = bytearray(5)
        msv = MyStruct.view(buffer, offset=1)
        msv.m1 = 1
        msv.m2 = 2
        assert pack(msv, MyStruct) == bytearray([0x01, 0x02, 0x00])

    def test_pack_kw_argument(self):
        """Test pack() utility supports view as keyword argument."""
        buffer = bytearray(5)
        msv = MyStruct.view(buffer, offset=1)
        msv.m1 = 1
        msv.m2 = 2
        assert pack({"x": msv}, {"x": MyStruct}) == bytearray([0x01, 0x02, 0x00])

    def test_get(self):
        """Test get() returns unpacked value."""
        buffer = bytearray([0x00, 0x01, 0x02, 0x00, 0xFF])
        msv = MyStruct.view(buffer, offset=1)
        value = msv.unpack()
        assert isinstance(value, MyStruct)
        assert value == [1, 2]
        # changing buffer does not effect previously returned value
        buffer[0:5] = [0] * 5
        assert value == [1, 2]

    def test_set(self):
        """Test set()."""
        buffer = bytearray(5)
        msv = MyStruct.view(buffer, offset=1)
        assert msv == [0, 0]
        msv.set([1, 2])
        assert msv == [1, 2]
        assert buffer == bytearray([0x00, 0x01, 0x02, 0x00, 0x00])

    def test_setattr(self):
        """Test __setattr__() method."""
        view = MyStruct.view(bytearray(5))
        view.abc = 1
        assert view.abc == 1


class TestBlockedMethods:

    """Test structure mutation methods are blocked."""

    def test_append(self):
        """Test append() method blocked."""
        view = MyStruct.view(bytearray(5))

        with pytest.raises(TypeError) as trap:
            view.append(0)

        expected = Baseline(
            """
            'StructureView' does not support append()
            """
        )

        assert wrap_message(trap.value) == expected

    def test_clear(self):
        """Test clear() method blocked."""
        view = MyStruct.view(bytearray(5))

        with pytest.raises(TypeError) as trap:
            view.clear()

        expected = Baseline(
            """
            'StructureView' does not support clear()
            """
        )

        assert wrap_message(trap.value) == expected

    def test_delattr(self):
        """Test __delattr__() method blocked."""
        view = MyStruct.view(bytearray(5))

        with pytest.raises(TypeError) as trap:
            del view.abc

        expected = Baseline(
            """
            'StructureView' does not support attribute deletion
            """
        )

        assert wrap_message(trap.value) == expected

    def test_delitem(self):
        """Test __delitem__() method blocked."""
        view = MyStruct.view(bytearray(5))
        with pytest.raises(TypeError) as trap:
            del view[0]

        expected = Baseline(
            """
            'StructureView' does not support item deletion
            """
        )

        assert wrap_message(trap.value) == expected

    def test_extend(self):
        """Test extend() method blocked."""
        view = MyStruct.view(bytearray(5))

        with pytest.raises(TypeError) as trap:
            view.extend([0])

        expected = Baseline(
            """
            'StructureView' does not support extend()
            """
        )

        assert wrap_message(trap.value) == expected

    def test_getattr(self):
        """Test __getattr__() method only allows access to structure members."""
        view = MyStruct.view(bytearray(5))

        with pytest.raises(AttributeError) as trap:
            view.abc  # pylint: disable=pointless-statement

        expected = Baseline(
            """
            'StructureView' object has no attribute 'abc'
            """
        )

        assert wrap_message(trap.value) == expected

    def test_iadd(self):
        """Test __iadd__() method blocked."""
        view = MyStruct.view(bytearray(5))

        with pytest.raises(TypeError) as trap:
            view += [0]

        expected = Baseline(
            """
            'StructureView' does not support in-place addition
            """
        )

        assert wrap_message(trap.value) == expected

    def test_imul(self):
        """Test __imul__() method blocked."""
        view = MyStruct.view(bytearray(5))

        with pytest.raises(TypeError) as trap:
            view *= [0]

        expected = Baseline(
            """
            'StructureView' does not support in-place multiplication
            """
        )

        assert wrap_message(trap.value) == expected

    def test_insert(self):
        """Test insert() method blocked."""
        view = MyStruct.view(bytearray(5))

        with pytest.raises(TypeError) as trap:
            view.insert(0, 0)

        expected = Baseline(
            """
            'StructureView' does not support insert()
            """
        )

        assert wrap_message(trap.value) == expected

    def test_pop(self):
        """Test pop() method blocked."""
        view = MyStruct.view(bytearray(5))

        with pytest.raises(TypeError) as trap:
            view.pop(0)

        expected = Baseline(
            """
            'StructureView' does not support pop()
            """
        )

        assert wrap_message(trap.value) == expected

    def test_remove(self):
        """Test remove() method blocked."""
        view = MyStruct.view(bytearray(5))

        with pytest.raises(TypeError) as trap:
            view.remove(0)

        expected = Baseline(
            """
            'StructureView' does not support remove()
            """
        )

        assert wrap_message(trap.value) == expected

    def test_reverse(self):
        """Test reverse() method blocked."""
        view = MyStruct.view(bytearray(5))

        with pytest.raises(TypeError) as trap:
            view.reverse()

        expected = Baseline(
            """
            'StructureView' does not support reverse()
            """
        )

        assert wrap_message(trap.value) == expected



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test NoneX transform properties."""

from plum.none import NoneX


class TestPositional:

    """Test explicitly defined with positional argument."""

    nonex = NoneX("nonex")

    def test_name(self):
        assert self.nonex.name == "nonex"

    def test_nbytes(self):
        assert self.nonex.nbytes == 0


class TestKeyword(TestPositional):

    """Test explicitly defined with keyword argument."""

    nonex = NoneX(name="nonex")



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Bytes transform for values that may be optionally present."""

from typing import Any, List, Optional, Tuple

from .dump import Record
from .transform import Transform


class OptionalX(Transform):

    """Bytes transform for values that may be optionally present."""

    def __init__(self, *, fmt: Any, name: Optional[str] = None) -> None:
        if name is None:
            if isinstance(fmt, Transform):
                name = f"Optional[{fmt.name}]"
            else:
                # must be a structure or bitfields
                name = f"Optional[{fmt.__name__}]"

        super().__init__(name)

        self._fmt = fmt

    @property
    def fmt(self) -> Any:
        """Optional format."""
        return self._fmt

    def __pack__(
        self, value: None, pieces: List[bytes], dump: Optional[Record] = None
    ) -> None:
        if value is None:
            if dump is not None:
                dump.value = repr(value)

        else:
            self._fmt.__pack__(value, pieces, dump)

    def __unpack__(
        self, buffer: bytes, offset: int, dump: Optional[Record] = None
    ) -> Tuple[None, int]:
        if buffer[offset:]:
            item, offset = self._fmt.__unpack__(buffer, offset, dump)

        else:
            item = None

            if dump is not None:
                dump.value = "None"

        return item, offset



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test Structure type read-only feature."""

import sys

import pytest
from baseline import Baseline

from plum.conformance import extract_method_code, wrap_message
from plum.exceptions import UnpackError
from plum.littleendian import uint8
from plum.structure import Structure, bitfield_member, member


class TestNormal:

    """Test readonly feature, normal members."""

    class Custom(Structure):
        """Sample structure type."""

        getable: int = member(fmt=uint8, readonly=True)
        setable: int = member(fmt=uint8, default=2)
        constant: int = member(fmt=uint8, default=3, readonly=True)

    def test_set_access_allowed(self):
        custom = self.Custom(getable=1, setable=2, constant=3)
        assert custom.setable == 2
        custom.setable = 0
        assert custom.setable == 0

    def test_set_access_blocked(self):
        expected_message = Baseline(
            """
            can't set attribute 'getable'
            """
        )

        custom = self.Custom(getable=1, setable=2, constant=3)

        with pytest.raises(AttributeError) as trap:
            custom.getable = 0

        actual_message = str(trap.value)

        if sys.version_info < (3, 10):
            actual_message += " 'getable'"

        assert actual_message == expected_message

    expected_message = Baseline(
        """
        +--------+----------+-------+-------+--------+
        | Offset | Access   | Value | Bytes | Format |
        +--------+----------+-------+-------+--------+
        |        |          |       |       | Custom |
        | 0      | getable  | 0     | 00    | uint8  |
        | 1      | setable  | 0     | 00    | uint8  |
        | 2      | constant | 0     | 00    | uint8  |
        +--------+----------+-------+-------+--------+

        ValueError occurred during unpack operation:

        'constant' must be 3
        """
    )

    def test_unpack_constant_check(self):
        with pytest.raises(UnpackError) as trap:
            self.Custom.unpack(bytes(3))

        assert wrap_message(trap.value) == self.expected_message

    implementation_baseline = Baseline(
        """
        def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["Custom", int]:
            structure = list.__new__(cls)

            if dump is None:
                m_getable, offset = uint8.__unpack__(buffer, offset, dump)

                m_setable, offset = uint8.__unpack__(buffer, offset, dump)

                m_constant, offset = uint8.__unpack__(buffer, offset, dump)
                if m_constant != 3:
                    raise ValueError(f"'constant' must be {3}")

            else:
                getable_dump = dump.add_record(access="getable", fmt=uint8)
                m_getable, offset = uint8.__unpack__(buffer, offset, getable_dump)

                setable_dump = dump.add_record(access="setable", fmt=uint8)
                m_setable, offset = uint8.__unpack__(buffer, offset, setable_dump)

                constant_dump = dump.add_record(access="constant", fmt=uint8)
                m_constant, offset = uint8.__unpack__(buffer, offset, constant_dump)
                if m_constant != 3:
                    raise ValueError(f"'constant' must be {3}")

            structure[:] = (m_getable, m_setable, m_constant)

            return structure, offset

        """
    )

    def test_unpack_implementation(self):
        unpack_implementation = extract_method_code(
            self.Custom.implementation, "__unpack__"
        )
        assert unpack_implementation == self.implementation_baseline


class TestBitField(TestNormal):

    """Test readonly feature, normal members."""

    class Custom(Structure, fieldorder="least_to_most"):
        """Sample structure type."""

        getable: int = bitfield_member(size=4, readonly=True)
        setable: int = bitfield_member(size=4, default=2)
        constant: int = bitfield_member(size=4, default=3, readonly=True)

    expected_message = Baseline(
        """
        +---------+----------+-------+-------------------+--------+
        | Offset  | Access   | Value | Bytes             | Format |
        +---------+----------+-------+-------------------+--------+
        |         |          |       |                   | Custom |
        | 0       |          | 0     | 00 00             |        |
        |  [0:4]  | getable  | 0     | ........ ....0000 | int    |
        |  [4:8]  | setable  | 0     | ........ 0000.... | int    |
        |  [8:12] | constant | 0     | ....0000 ........ | int    |
        +---------+----------+-------+-------------------+--------+

        ValueError occurred during unpack operation:

        'constant' must be 3
        """
    )

    implementation_baseline = Baseline(
        """
        def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["Custom", int]:
            structure = list.__new__(cls)

            if dump is None:
                bitfields_0, offset = cls.getable.__unpack__(buffer, offset, dump, "little", nbytes=2, signed=False)

                m_getable = (bitfields_0 >> 0) & 0xf

                m_setable = (bitfields_0 >> 4) & 0xf

                m_constant = (bitfields_0 >> 8) & 0xf
                if m_constant != 3:
                    raise ValueError(f"'constant' must be {3}")

            else:
                bitfields_0_dump = dump.add_record()
                bitfields_0, offset = cls.getable.__unpack__(buffer, offset, bitfields_0_dump, "little", nbytes=2, signed=False)

                m_getable = (bitfields_0 >> 0) & 0xf
                dump.add_record(access="getable", bits=(0, 4), value=m_getable, fmt=int)

                m_setable = (bitfields_0 >> 4) & 0xf
                dump.add_record(access="setable", bits=(4, 4), value=m_setable, fmt=int)

                m_constant = (bitfields_0 >> 8) & 0xf
                dump.add_record(access="constant", bits=(8, 4), value=m_constant, fmt=int)
                if m_constant != 3:
                    raise ValueError(f"'constant' must be {3}")

            structure[:] = (m_getable, m_setable, m_constant)

            return structure, offset

        """
    )



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test BitFields data store transform properties."""
# pylint: disable=comparison-with-callable

from plum.bitfields import BitFields, bitfield


class TestDefault:

    """Test with as many left to default as possible."""

    class Sample(BitFields):

        """Sample bit fields data store transform type."""

        m0: int = bitfield(size=1)
        m1: int = bitfield(size=1)

    def test_nbytes(self):
        assert self.Sample.nbytes == 1
        assert self.Sample.from_int(0).nbytes == 1

    def test_byteorder(self):
        assert self.Sample.byteorder == "little"

    def test_default(self):
        assert self.Sample.default == 0

    def test_ignore(self):
        assert self.Sample.ignore == 0

    def test_nested(self):
        assert self.Sample.nested is False


class TestKeyword:

    """Test explicitly defined with keyword argument."""

    class Sample(
        BitFields, nbytes=2, byteorder="big", default=1, ignore=2, nested=True
    ):

        """Sample bit fields data store transform type."""

        m0: int = bitfield(size=1)
        m1: int = bitfield(size=1)

    def test_nbytes(self):
        assert self.Sample.nbytes == 2
        assert self.Sample.from_int(0).nbytes == 2

    def test_byteorder(self):
        assert self.Sample.byteorder == "big"

    def test_default(self):
        assert self.Sample.default == 1

    def test_ignore(self):
        assert self.Sample.ignore == 2

    def test_nested(self):
        assert self.Sample.nested is True



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test float transform properties."""

from plum.float import FloatX


class TestDefault:

    """Test with as many left to default as possible."""

    float32 = FloatX("float32", nbytes=4)

    def test_name(self):
        assert self.float32.name == "float32"

    def test_nbytes(self):
        assert self.float32.nbytes == 4

    def test_byteorder(self):
        assert self.float32.byteorder == "little"


class TestPositional:

    """Test explicitly defined with positional argument."""

    float32 = FloatX("float32", 4, "big")

    def test_name(self):
        assert self.float32.name == "float32"

    def test_nbytes(self):
        assert self.float32.nbytes == 4

    def test_byteorder(self):
        assert self.float32.byteorder == "big"


class TestKeyword(TestPositional):

    """Test explicitly defined with keyword argument."""

    float32 = FloatX(name="float32", nbytes=4, byteorder="big")



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test integer enumeration classes from little endian byte order module."""

import pytest
from baseline import Baseline

from plum.conformance import wrap_message
from plum.enum import EnumX
from plum.exceptions import PackError
from plum.utilities import pack

from sample_enum import Register


register16 = EnumX("register16", Register, nbytes=2, byteorder="big", signed=False)


class TestPack:

    """Test pack exceptions."""

    def test_not_int_val(self):
        """Test not integer values."""
        with pytest.raises(PackError) as trap:
            pack("str", register16)

        expected = Baseline(
            """
            +--------+-------+-------+------------+
            | Offset | Value | Bytes | Format     |
            +--------+-------+-------+------------+
            |        | 'str' |       | register16 |
            +--------+-------+-------+------------+

            ValueError occurred during pack operation:

            'str' is not a valid Register
            """
        )

        assert wrap_message(trap.value) == expected



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test basic features of BitField member class."""

import pytest
from baseline import Baseline

from plum.bitfields import BitFields, bitfield
from plum.conformance import wrap_message

from sample_bitfields import MyBits


class TestFeatures:

    """Test misc. features."""

    def test_repr(self):
        expected = Baseline(
            """
            BitField(name='bitfield')
            """
        )

        bitfield_ = bitfield(lsb=0, size=1)
        bitfield_.finish_initialization("bitfield")

        assert repr(bitfield_) == expected

    def test_default_type(self):
        # no arg, no annotation
        bitfield_ = bitfield(lsb=0, size=1)
        bitfield_.finish_initialization("bitfield")
        assert bitfield_.typ is int


class TestInvalidPosition:

    """Test invalid bit position for bit field member."""

    def test_less_than_zero(self):
        """Test bit position less than zero."""

        expected = Baseline(
            """
            bit field position must be greater than or equal to zero
            """
        )

        with pytest.raises(TypeError) as trap:
            bitfield(lsb=-1, size=2, signed=True)

        assert wrap_message(trap.value) == expected


class TestInvalidTypExplicit:

    """Test invalid bit field typ when specified incorrectly in member."""

    def test_instance(self):
        """Test typ is not a class but an instance of a class."""

        expected = Baseline(
            """
            bit field type must be int-like
            """
        )

        with pytest.raises(TypeError) as trap:
            bitfield(typ=1, lsb=0, size=2, signed=True)

        assert wrap_message(trap.value) == expected

    def test_invalid_class(self):
        """Test typ is not int-like."""

        expected = Baseline(
            """
            bit field type must be int-like
            """
        )

        with pytest.raises(TypeError) as trap:
            bitfield(typ=float, lsb=0, size=2, signed=True)

        assert wrap_message(trap.value) == expected

    def test_not_marked_nested(self):
        """Test typ as a BitFields but wasn't set up to support nested."""

        expected = Baseline(
            """
            bit field typ must be declared as nested (e.g. 'class
            MyBits(BitFields, nested=True):')
            """
        )

        with pytest.raises(TypeError) as trap:
            bitfield(size=2, typ=MyBits)

        assert wrap_message(trap.value) == expected


class TestInvalidTypAnnotation:

    """Test invalid bit field typ when specified via annotation."""

    def test_instance(self):
        """Test typ is not a class but an instance of a class."""

        expected = Baseline(
            """
            bit field type must be int-like
            """
        )

        with pytest.raises(TypeError) as trap:

            class MyBitFields(BitFields):  # pylint: disable=unused-variable
                f1 = bitfield(lsb=0, size=2, signed=True, typ=1)

        assert wrap_message(trap.value) == expected

    def test_invalid_class(self):
        """Test typ is not int-like."""

        expected = Baseline(
            """
            bit field type must be int-like
            """
        )

        with pytest.raises(TypeError) as trap:

            class MyBitFields(BitFields):  # pylint: disable=unused-variable
                f1 = bitfield(lsb=0, size=2, signed=True, typ=float)

        assert wrap_message(trap.value) == expected


class TestInvalidSizes:

    """Test invalid bit sizes for bit field member."""

    def test_signed_int(self):
        """Test bit size restriction for signed int."""

        # no exception
        bitfield(lsb=0, size=2, signed=True)

        expected = Baseline(
            """
            'size' must be 2 or greater for signed bit field
            """
        )

        with pytest.raises(ValueError) as trap:
            bitfield(lsb=0, size=1, signed=True)
        assert wrap_message(trap.value) == expected

        with pytest.raises(ValueError) as trap:
            bitfield(lsb=0, size=0, signed=True)
        assert wrap_message(trap.value) == expected

        with pytest.raises(ValueError) as trap:
            bitfield(lsb=0, size=-1, signed=True)
        assert wrap_message(trap.value) == expected

    def test_unsigned_int(self):
        """Test bit size restriction for unsigned int."""

        # no exception
        bitfield(lsb=0, size=1, signed=False)

        expected = Baseline(
            """
            'size' must be 1 or greater for unsigned bit field
            """
        )

        with pytest.raises(ValueError) as trap:
            bitfield(lsb=0, size=0, signed=False)
        assert wrap_message(trap.value) == expected

        with pytest.raises(ValueError) as trap:
            bitfield(lsb=0, size=-1, signed=False)
        assert wrap_message(trap.value) == expected


class TestInvalidDefault:

    """Test out of range default for bit field member."""

    def test_too_small(self):
        expected = Baseline(
            """
            bit field requires 0 <= default <= 3
            """
        )

        with pytest.raises(ValueError) as trap:
            bitfield(size=2, default=-1)
        assert wrap_message(trap.value) == expected

    def test_too_big(self):
        expected = Baseline(
            """
            bit field requires 0 <= default <= 3
            """
        )

        with pytest.raises(ValueError) as trap:
            bitfield(size=2, default=4)
        assert wrap_message(trap.value) == expected



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test bytes transform support for structure's sized_member()."""

from plum.dump import Dump
from plum.bytes import BytesX


greedy_bytes = BytesX("greedy_bytes")


class TestUnpack:

    """Test unpack method variants."""

    buffer = b"\x00\x01\x02"

    def test_unpack(self):
        array, _offset = greedy_bytes.__unpack__(self.buffer, 0, nbytes=2)
        assert array == self.buffer[0:2]

    def test_unpack_and_dump(self):
        array, _offset = greedy_bytes.__unpack_and_dump__(
            self.buffer, 0, Dump(), nbytes=2
        )
        assert array == self.buffer[0:2]



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test IPv4AddressX transform properties."""

from plum.ipaddress import IPv4AddressX


class TestDefault:

    """Test explicitly defined with positional argument."""

    ipv4address = IPv4AddressX("ipv4address")

    def test_name(self):
        assert self.ipv4address.name == "ipv4address"

    def test_byteorder(self):
        assert self.ipv4address.byteorder == "big"

    def test_nbytes(self):
        assert self.ipv4address.nbytes == 4


class TestPositional:

    """Test explicitly defined with positional argument."""

    ipv4address = IPv4AddressX("ipv4address", "little")

    def test_name(self):
        assert self.ipv4address.name == "ipv4address"

    def test_byteorder(self):
        assert self.ipv4address.byteorder == "little"

    def test_nbytes(self):
        assert self.ipv4address.nbytes == 4


class TestKeyword(TestPositional):

    """Test explicitly defined with keyword argument."""

    ipv4address = IPv4AddressX(name="ipv4address", byteorder="little")



"""Decimal to bytes and bytes to decimal transform."""

from decimal import Decimal, InvalidOperation
from typing import Optional, Union, Type, Tuple, List
from plum.transform import Transform
from plum.data import Data
from plum.dump import Record
from plum.exceptions import InsufficientMemoryError
from plum._getbytes import getbytes
from ._typing import ByteOrderHint

DREF = Optional[Union[Type[Data], Transform]]


class DecimalX(Transform):

    """Decimal to bytes and bytes to decimal transform."""

    __dref__: DREF
    __nbytes__: int
    __bdns__: Tuple[ByteOrderHint, int, int, bool]
    __precision__: int

    def __init__(
        self,
        name: str,
        nbytes: int,
        precision: int,
        byteorder: ByteOrderHint = "little",
        *,
        signed: bool = False,
        dref: DREF = None,
    ) -> None:
        super().__init__(name)

        assert nbytes > 0
        assert byteorder in {"big", "little"}
        assert precision >= 0

        self.__dref__ = dref
        self.__nbytes__ = nbytes
        divisor = 10**precision
        self.__bdns__ = byteorder, divisor, nbytes, signed
        self.__precision__ = precision

    @property
    def byteorder(self) -> ByteOrderHint:
        """Byte order ("little" or "big")."""
        return self.__bdns__[0]

    @property
    def signed(self) -> bool:
        """Signed decimal."""
        return self.__bdns__[3]

    @property
    def precision(self) -> int:
        """Precision."""
        return self.__precision__

    def __pack__(
        self,
        value: Union[int, float, str, Decimal],
        pieces: List[bytes],
        dump: Optional[Record] = None,
    ) -> None:
        byteorder, divisor, nbytes, signed = self.__bdns__

        try:
            decimal_value = Decimal(value)
        except InvalidOperation:
            raise ValueError(
                f"Value {value!r} cannot be converted to decimal."
            ) from None

        decimal_value *= divisor
        int_value = int(decimal_value.to_integral_value())
        if dump is None:
            pieces.append(int_value.to_bytes(nbytes, byteorder, signed=signed))
        else:
            dump.value = repr(value)

            piece = int_value.to_bytes(nbytes, byteorder, signed=signed)

            dump.value = value
            dump.memory = piece

            pieces.append(piece)

    def __unpack__(
        self, buffer: bytes, offset: int, dump: Optional[Record] = None
    ) -> Tuple[Decimal, int]:
        byteorder, divisor, nbytes, signed = self.__bdns__

        if dump is None:
            end = offset + nbytes

            if len(buffer) < end:
                raise InsufficientMemoryError("too few bytes to unpack")

            int_value = int.from_bytes(buffer[offset:end], byteorder, signed=signed)
            decimal_value = Decimal(int_value) / divisor
            return decimal_value, end

        chunk, end = getbytes(buffer, offset, dump, nbytes)

        int_value = int.from_bytes(chunk, byteorder, signed=signed)

        decimal_value = Decimal(int_value) / divisor
        dump.value = decimal_value
        return decimal_value, end



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test Data store base class."""

# anomaly in pylint: disable=no-member

import pytest
from baseline import Baseline

from plum.conformance import wrap_message
from plum.data import Data
from plum.exceptions import ExcessMemoryError, PackError, UnpackError
from plum.littleendian import uint16


class TestRetries:

    """Test pack and unpack error retry and reporting mechanisms."""

    class Custom(Data):
        __nbytes__ = 2

        def __init__(self, value=0):
            self.value = value

        def __int__(self):
            return int(self.value)

        @classmethod
        def __pack__(cls, value, pieces, dump=None):
            if dump is not None:
                dump.value = value
            uint16.__pack__(int(value), pieces, dump)

        @classmethod
        def __unpack__(cls, buffer, offset, dump=None):
            value, offset = uint16.__unpack__(buffer, offset, dump)
            return cls(value), offset

        def __eq__(self, other):
            return int(self) == int(other)

        def __repr__(self):
            return f"{self.value}"

    def test_unpack_error(self):
        """Exercise formatting all of the excessive bytes."""
        expected_message = Baseline(
            """
            +--------+----------------+-------------------------------------------------+--------+
            | Offset | Value          | Bytes                                           | Format |
            +--------+----------------+-------------------------------------------------+--------+
            |  0     | 39321          | 99 99                                           | Custom |
            +--------+----------------+-------------------------------------------------+--------+
            |  2     | <excess bytes> | 99 99 99 99 99 99 99 99 99 99 99 99 99 99 99 99 |        |
            | 18     |                | 99 99 99 99                                     |        |
            +--------+----------------+-------------------------------------------------+--------+

            ExcessMemoryError occurred during unpack operation:

            20 unconsumed bytes
            """
        )
        with pytest.raises(UnpackError) as trap:
            self.Custom.unpack(b"\x99" * 22)

        assert wrap_message(trap.value) == expected_message
        assert isinstance(trap.value.__context__, ExcessMemoryError)

    expected_pack_error = Baseline(
        """
        +--------+-------+-------+--------+
        | Offset | Value | Bytes | Format |
        +--------+-------+-------+--------+
        |        | hello |       | Custom |
        +--------+-------+-------+--------+

        ValueError occurred during pack operation:

        invalid literal for int() with base 10: 'hello'
        """
    )

    def test_pack_error(self):
        with pytest.raises(PackError) as trap:
            self.Custom.pack("hello")

        assert wrap_message(trap.value) == self.expected_pack_error

    def test_ipack_error(self):
        with pytest.raises(PackError) as trap:
            self.Custom("hello").ipack()

        assert wrap_message(trap.value) == self.expected_pack_error

    def test_view(self):
        expected_message = Baseline(
            """
            'Custom' does not support view()
            """
        )
        with pytest.raises(TypeError) as trap:
            self.Custom.view(bytes(2))

        assert wrap_message(trap.value) == expected_message



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test exception corner cases with Array type usage."""

import pytest
from baseline import Baseline

from plum.array import ArrayX
from plum.bigendian import uint8
from plum.conformance import wrap_message
from plum.exceptions import PackError, UnpackError
from plum.utilities import pack, unpack


class TestUnpack:

    """Test exception cases during unpack."""

    def test_greedy_multidimensional(self):
        """Verify exception/message when None specified for one of the dimensions."""

        # sample array type with second dimension unspecified
        customarray = ArrayX("customarray", fmt=uint8, dims=(2, None))

        with pytest.raises(UnpackError) as trap:
            unpack(customarray, bytes(10))

        expected = Baseline(
            """
            +--------+-------+-------+-------------+
            | Offset | Value | Bytes | Format      |
            +--------+-------+-------+-------------+
            |        |       |       | customarray |
            +--------+-------+-------+-------------+

            TypeError occurred during unpack operation:

            array unpack does not support greedy secondary dimensions
            """
        )

        assert wrap_message(trap.value) == expected


class TestPack:

    """Test exception cases during pack."""

    def test_greedy_noniterable(self):
        """Pack non-iterable value with greedy array."""

        array = ArrayX("array", fmt=uint8)

        with pytest.raises(PackError) as trap:
            pack(0, array)

        expected = Baseline(
            """
            +--------+-------+-------+--------+
            | Offset | Value | Bytes | Format |
            +--------+-------+-------+--------+
            |        | 0     |       | array  |
            +--------+-------+-------+--------+

            TypeError occurred during pack operation:

            invalid value, expected iterable of any length, got non-iterable
            """
        )

        assert wrap_message(trap.value) == expected

    def test_dimensioned_noniterable(self):
        """Pack non-iterable value with fixed dimension array."""
        customarray = ArrayX("customarray", fmt=uint8, dims=(2,))

        with pytest.raises(PackError) as trap:
            pack(0, customarray)

        expected = Baseline(
            """
            +--------+-------+-------+-------------+
            | Offset | Value | Bytes | Format      |
            +--------+-------+-------+-------------+
            |        | 0     |       | customarray |
            +--------+-------+-------+-------------+

            TypeError occurred during pack operation:

            invalid value, expected iterable of length 2, got non-iterable
            """
        )

        assert wrap_message(trap.value) == expected

    def test_dimension_mismatch_missing(self):
        """Pack iterable value but length does not match array fixed dimension (missing)."""
        customarray = ArrayX("customarray", fmt=uint8, dims=(2,))

        with pytest.raises(PackError) as trap:
            pack([1], customarray)

        expected = Baseline(
            """
            +--------+--------+-----------+-------+-------------+
            | Offset | Access | Value     | Bytes | Format      |
            +--------+--------+-----------+-------+-------------+
            |        |        |           |       | customarray |
            | 0      | [0]    | 1         | 01    | uint8       |
            |        | [1]    | <missing> |       |             |
            +--------+--------+-----------+-------+-------------+

            TypeError occurred during pack operation:

            invalid value, expected iterable of 2 length, got iterable of length 1
            """
        )

        assert wrap_message(trap.value) == expected

    def test_dimension_mismatch_extra(self):
        """Pack iterable value but length does not match array fixed dimension (extra)."""
        customarray = ArrayX("customarray", fmt=uint8, dims=(2,))

        with pytest.raises(PackError) as trap:
            pack([1, 2, 3], customarray)

        expected = Baseline(
            """
            +--------+-------------+-------+-------+-------------+
            | Offset | Access      | Value | Bytes | Format      |
            +--------+-------------+-------+-------+-------------+
            |        |             |       |       | customarray |
            | 0      | [0]         | 1     | 01    | uint8       |
            | 1      | [1]         | 2     | 02    | uint8       |
            +--------+-------------+-------+-------+-------------+
            |        | [2] <extra> | 3     |       |             |
            +--------+-------------+-------+-------+-------------+

            TypeError occurred during pack operation:

            invalid value, expected iterable of 2 length, got iterable of length 3
            """
        )

        assert wrap_message(trap.value) == expected



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test SizedX transform API conformance."""

import pytest

from baseline import Baseline

from plum.conformance import wrap_message
from plum.exceptions import UnpackError
from plum.littleendian import uint8
from plum.sized import SizedX
from plum.str import StrX


ascii_greedy = StrX("ascii", encoding="ascii")

sized_string = SizedX(
    "sized_string", fmt=ascii_greedy, size_fmt=uint8, size_access="--nbytes--"
)


class TestSizeAccess:

    """Test customization of size access description."""

    expected_dump = Baseline(
        """
        +--------+------------+----------------+-------------------------------------+--------------+
        | Offset | Access     | Value          | Bytes                               | Format       |
        +--------+------------+----------------+-------------------------------------+--------------+
        |        |            |                |                                     | sized_string |
        |  0     | --nbytes-- | 12             | 0c                                  | uint8        |
        |        |            |                |                                     | ascii        |
        |  1     | [0:12]     | 'Hello World!' | 48 65 6c 6c 6f 20 57 6f 72 6c 64 21 |              |
        +--------+------------+----------------+-------------------------------------+--------------+
        """
    )

    bindata = b"\x0cHello World!"

    def test_pack_dump(self):
        _bindata, dump = sized_string.pack_and_dump("Hello World!")
        assert str(dump) == self.expected_dump

    def test_unpack_dump(self):
        _value, dump = sized_string.unpack_and_dump(self.bindata)
        assert str(dump) == self.expected_dump


class TestExceptions:

    """Test special exception cases."""

    def test_excess_bytes(self):
        """Test outer limit more restrictive than inner limit (string case)."""
        sized_int = SizedX("sized_string", fmt=uint8, size_fmt=uint8)

        with pytest.raises(UnpackError) as trap:
            sized_int.unpack(b"\x20" + bytes(32))

        expected = Baseline(
            """
            +--------+----------+----------------+-------------------------------------------------+--------------+
            | Offset | Access   | Value          | Bytes                                           | Format       |
            +--------+----------+----------------+-------------------------------------------------+--------------+
            |        |          |                |                                                 | sized_string |
            |  0     | --size-- | 32             | 20                                              | uint8        |
            |  1     |          | 0              | 00                                              | uint8        |
            +--------+----------+----------------+-------------------------------------------------+--------------+
            |  2     |          | <excess bytes> | 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 |              |
            | 18     |          |                | 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00    |              |
            +--------+----------+----------------+-------------------------------------------------+--------------+

            ExcessMemoryError occurred during unpack operation:

            31 unconsumed bytes
            """
        )

        assert wrap_message(trap.value) == expected


class TestSizedProperties:

    """Test SizedX properties."""

    def test_fmt(self):
        assert sized_string.fmt is ascii_greedy

    def test_offset(self):
        assert sized_string.offset == 0

    def test_ratio(self):
        assert sized_string.ratio == 1

    def test_size_access(self):
        assert sized_string.size_access == "--nbytes--"

    def test_size_fmt(self):
        assert sized_string.size_fmt is uint8



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Common transforms in big endian format."""

from .float import FloatX
from .int import IntX


sint8 = IntX("sint8", nbytes=1, byteorder="big", signed=True)
sint16 = IntX("sint16", nbytes=2, byteorder="big", signed=True)
sint32 = IntX("sint32", nbytes=4, byteorder="big", signed=True)
sint64 = IntX("sint64", nbytes=8, byteorder="big", signed=True)

uint8 = IntX("uint8", nbytes=1, byteorder="big", signed=False)
uint16 = IntX("uint16", nbytes=2, byteorder="big", signed=False)
uint32 = IntX("uint32", nbytes=4, byteorder="big", signed=False)
uint64 = IntX("uint64", nbytes=8, byteorder="big", signed=False)

single = FloatX("single", nbytes=4, byteorder="big")
double = FloatX("double", nbytes=8, byteorder="big")



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test default factory function as default."""

from baseline import Baseline

from plum.conformance import Case, CaseData
from plum.littleendian import uint8
from plum.structure import Structure, member


class TestMethod(Case):

    """Factory passed half baked structure instance (factory is a simple method)."""

    class Struct(Structure):
        def data_default(self):
            mapping = {0: 0x10, 1: 0x20}
            return mapping[self[0]]

        prev: int = member(fmt=uint8)
        data: int = member(fmt=uint8, default=data_default)

    data = CaseData(
        fmt=Struct,
        bindata=bytes.fromhex("0010"),
        nbytes=2,
        values=(
            Struct(prev=0, data=0x10),
            Struct(prev=0),
            [0, 0x10],
        ),
        dump=Baseline(
            """
            +--------+--------+-------+-------+--------+
            | Offset | Access | Value | Bytes | Format |
            +--------+--------+-------+-------+--------+
            |        |        |       |       | Struct |
            | 0      | prev   | 0     | 00    | uint8  |
            | 1      | data   | 16    | 10    | uint8  |
            +--------+--------+-------+-------+--------+
            """
        ),
        excess=Baseline(
            """
            +--------+--------+----------------+-------+--------+
            | Offset | Access | Value          | Bytes | Format |
            +--------+--------+----------------+-------+--------+
            |        |        |                |       | Struct |
            | 0      | prev   | 0              | 00    | uint8  |
            | 1      | data   | 16             | 10    | uint8  |
            +--------+--------+----------------+-------+--------+
            | 2      |        | <excess bytes> | 99    |        |
            +--------+--------+----------------+-------+--------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
            +--------+--------+----------------------+-------+--------+
            | Offset | Access | Value                | Bytes | Format |
            +--------+--------+----------------------+-------+--------+
            |        |        |                      |       | Struct |
            | 0      | prev   | 0                    | 00    | uint8  |
            |        | data   | <insufficient bytes> |       | uint8  |
            +--------+--------+----------------------+-------+--------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack 'uint8' (1 needed, only 0 available)
            """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, prev: int, data: Optional[int] = None) -> None:
                self[:] = (prev, data)

                if data is None:
                    self[1] = type(self).data.default(self)

            @prev.getter
            def prev(self) -> int:
                return self[0]

            @prev.setter
            def prev(self, value: int) -> None:
                self[0] = value

            @data.getter
            def data(self) -> int:
                return self[1]

            @data.setter
            def data(self, value: int) -> None:
                self[1] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_prev, m_data) = value

                if dump is None:
                    uint8.__pack__(m_prev, pieces, dump)

                    uint8.__pack__(m_data, pieces, dump)

                else:
                    prev_dump = dump.add_record(access="prev", fmt=uint8)
                    uint8.__pack__(m_prev, pieces, prev_dump)

                    data_dump = dump.add_record(access="data", fmt=uint8)
                    uint8.__pack__(m_data, pieces, data_dump)

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["Struct", int]:
                structure = list.__new__(cls)

                if dump is None:
                    m_prev, offset = uint8.__unpack__(buffer, offset, dump)

                    m_data, offset = uint8.__unpack__(buffer, offset, dump)

                else:
                    prev_dump = dump.add_record(access="prev", fmt=uint8)
                    m_prev, offset = uint8.__unpack__(buffer, offset, prev_dump)

                    data_dump = dump.add_record(access="data", fmt=uint8)
                    m_data, offset = uint8.__unpack__(buffer, offset, data_dump)

                structure[:] = (m_prev, m_data)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(prev={self.prev!r}, data={self.data!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )


class TestProperty(Case):

    """Factory passed half baked structure instance (factory is a property method)."""

    class Struct(Structure):
        @property
        def data_default(self):
            mapping = {0: 0x10, 1: 0x20}
            return mapping[self[0]]

        prev: int = member(fmt=uint8)
        data: int = member(fmt=uint8, default=data_default)

    data = CaseData(
        fmt=Struct,
        bindata=bytes.fromhex("0010"),
        nbytes=2,
        values=(
            Struct(prev=0, data=0x10),
            Struct(prev=0),
            [0, 0x10],
        ),
        dump=Baseline(
            """
            +--------+--------+-------+-------+--------+
            | Offset | Access | Value | Bytes | Format |
            +--------+--------+-------+-------+--------+
            |        |        |       |       | Struct |
            | 0      | prev   | 0     | 00    | uint8  |
            | 1      | data   | 16    | 10    | uint8  |
            +--------+--------+-------+-------+--------+
            """
        ),
        excess=Baseline(
            """
            +--------+--------+----------------+-------+--------+
            | Offset | Access | Value          | Bytes | Format |
            +--------+--------+----------------+-------+--------+
            |        |        |                |       | Struct |
            | 0      | prev   | 0              | 00    | uint8  |
            | 1      | data   | 16             | 10    | uint8  |
            +--------+--------+----------------+-------+--------+
            | 2      |        | <excess bytes> | 99    |        |
            +--------+--------+----------------+-------+--------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
            +--------+--------+----------------------+-------+--------+
            | Offset | Access | Value                | Bytes | Format |
            +--------+--------+----------------------+-------+--------+
            |        |        |                      |       | Struct |
            | 0      | prev   | 0                    | 00    | uint8  |
            |        | data   | <insufficient bytes> |       | uint8  |
            +--------+--------+----------------------+-------+--------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack 'uint8' (1 needed, only 0 available)
            """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, prev: int, data: Optional[int] = None) -> None:
                self[:] = (prev, data)

                if data is None:
                    self[1] = type(self).data.default(self)

            @prev.getter
            def prev(self) -> int:
                return self[0]

            @prev.setter
            def prev(self, value: int) -> None:
                self[0] = value

            @data.getter
            def data(self) -> int:
                return self[1]

            @data.setter
            def data(self, value: int) -> None:
                self[1] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_prev, m_data) = value

                if dump is None:
                    uint8.__pack__(m_prev, pieces, dump)

                    uint8.__pack__(m_data, pieces, dump)

                else:
                    prev_dump = dump.add_record(access="prev", fmt=uint8)
                    uint8.__pack__(m_prev, pieces, prev_dump)

                    data_dump = dump.add_record(access="data", fmt=uint8)
                    uint8.__pack__(m_data, pieces, data_dump)

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["Struct", int]:
                structure = list.__new__(cls)

                if dump is None:
                    m_prev, offset = uint8.__unpack__(buffer, offset, dump)

                    m_data, offset = uint8.__unpack__(buffer, offset, dump)

                else:
                    prev_dump = dump.add_record(access="prev", fmt=uint8)
                    m_prev, offset = uint8.__unpack__(buffer, offset, prev_dump)

                    data_dump = dump.add_record(access="data", fmt=uint8)
                    m_data, offset = uint8.__unpack__(buffer, offset, data_dump)

                structure[:] = (m_prev, m_data)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(prev={self.prev!r}, data={self.data!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test integer transform properties."""

from plum.int import IntX


class TestDefault:

    """Test with as many left to default as possible."""

    uint8 = IntX("uint8", 1)

    def test_name(self):
        assert self.uint8.name == "uint8"

    def test_nbytes(self):
        assert self.uint8.nbytes == 1

    def test_byteorder(self):
        assert self.uint8.byteorder == "little"

    def test_signed(self):
        assert self.uint8.signed is False


class TestPositional:

    """Test explicitly defined with positional argument."""

    uint8 = IntX("uint8", 2, "big", signed=True)

    def test_name(self):
        assert self.uint8.name == "uint8"

    def test_nbytes(self):
        assert self.uint8.nbytes == 2

    def test_byteorder(self):
        assert self.uint8.byteorder == "big"

    def test_signed(self):
        assert self.uint8.signed is True


class TestKeyword(TestPositional):

    """Test explicitly defined with keyword argument."""

    uint8 = IntX(name="uint8", nbytes=2, byteorder="big", signed=True)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Variable dimensions array structure member definition."""

# pylint: disable=too-many-arguments

from typing import Any, Generator, List, Optional, Tuple

from .._code_injector import CodeInjector
from ..array import ArrayX
from ..int import IntX
from ._member import Member


class DimmedMember(Member):

    """Variably dimensioned array structure member definition.

    :param doc: accessor documentation string
    :param fmt: array transform (or factory function)
    :param dims: array dimensions member definition
    :param default: initializer default value
    :param ignore: ignore member during comparisons
    :param readonly: block setting member attribute
    :param argrepr: format to represent member argument in structure repr

    """

    _attributes_to_copy: Tuple[str, ...] = Member._attributes_to_copy + ("_dims",)

    def __init__(
        self,
        doc: str,
        fmt: ArrayX,
        dims: Member,
        default: Any,
        ignore: bool,
        readonly: bool,
        argrepr: Optional[str],
    ) -> None:
        compute = False
        fmt_arg = None
        super().__init__(
            doc,
            default,
            ignore,
            readonly,
            compute,
            argrepr,
            fmt,
            fmt_arg,
        )

        if not isinstance(dims, Member):
            raise TypeError("invalid 'dims', must be a 'member()'")

        self._dims: Member = dims

    @property
    def dims(self) -> Member:
        """Array dimensions member definition."""
        return self._dims

    @staticmethod
    def compute_dims(array: List[Any], ndims: int) -> Generator[int, None, None]:
        """Determine dimensions from multi-dimensional array value."""
        for _ in range(ndims):
            yield len(array)
            array = array[0]

    def add_name_index(
        self,
        name: str,
        index: int,
        code_injector: CodeInjector,
    ) -> int:
        """Assign name, index number, and prepare for code generation."""
        dims = self.dims

        dims.temp_store.num_associated += 1

        if dims.compute:

            # add logic to __init__ to compute dimensions when it was left to default
            if isinstance(dims.fmt, IntX):
                computation = f"len({name})"
            elif isinstance(dims.fmt, ArrayX):
                ndims = dims.fmt.__ndims__
                assert ndims, "dims member must be single dim and non-greedy"
                computation = f"list(type(self).{name}.compute_dims({name}, {ndims}))"
            else:
                raise TypeError(
                    f"{dims.name!r} dims member not an integer or array transform format"
                )

            dims.temp_store.init_begin += [
                f"if {dims.name} is None:",
                f"    {dims.name} = {computation}",
            ]

            # add logic to this member's setter to recompute associated dims member
            # (but only when setter not explicitly defined in structure class)
            if not self.readonly and not self.fset:
                if isinstance(self.dims.fmt, IntX):
                    computation = "len(value)"
                else:
                    num_dims = self.dims.fmt.__ndims__
                    computation = (
                        f"list(type(self).{name}.compute_dims(value, {num_dims}))"
                    )
                self.temp_store.setter.append(
                    f"self[{self.dims.index}] = {computation}  "
                    f"# update {self.dims.name!r} member"
                )

        return Member.add_name_index(self, name, index, code_injector)

    @property
    def extra_args(self) -> str:
        """Additional __pack__ and __unpack__ arguments (for code generation)."""
        if isinstance(self.dims.fmt, IntX):
            args = f", dims=(m_{self.dims.name}, )"
        else:
            args = f", dims=m_{self.dims.name}"

        return args


def dimmed_member(
    doc: str = "",
    *,
    fmt: ArrayX,
    dims: Member,
    default: Any = None,
    ignore: bool = False,
    readonly: bool = False,
    argrepr: Optional[str] = None,
) -> Any:
    """Define variably dimensioned array structure member.

    :param doc: accessor documentation string
    :param fmt: array transform (or factory function)
    :param dims: array dimensions member definition
    :param default: initializer default value
    :param ignore: ignore member during comparisons
    :param readonly: block setting member attribute
    :param argrepr: format to represent member argument in structure repr

    """
    return DimmedMember(doc, fmt, dims, default, ignore, readonly, argrepr)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test array transform properties."""

import pytest

from plum.array import ArrayX
from plum.exceptions import SizeError
from plum.littleendian import uint8
from plum.str import StrX

ascii_zero = StrX("ascii_zero", encoding="ascii", zero_termination=True)


class TestDefault:

    """Test with as many left to default as possible."""

    arrayx = ArrayX("arrayx", uint8)

    def test_name(self):
        assert self.arrayx.name == "arrayx"

    def test_nbytes(self):
        with pytest.raises(SizeError):
            return self.arrayx.nbytes

    def test_fmt(self):
        assert self.arrayx.fmt is uint8

    def test_dims(self):
        return self.arrayx.dims == (None,)


class TestPositional:

    """Test explicitly defined with positional argument."""

    arrayx = ArrayX("arrayx", uint8, (2,))

    def test_name(self):
        assert self.arrayx.name == "arrayx"

    def test_nbytes(self):
        assert self.arrayx.nbytes == 2

    def test_fmt(self):
        assert self.arrayx.fmt is uint8

    def test_dims(self):
        return self.arrayx.dims == (2,)


class TestKeyword(TestPositional):

    """Test explicitly defined with keyword argument."""

    arrayx = ArrayX(name="arrayx", fmt=uint8, dims=(2,))


class TestVariable:

    """Test nbytes property when member size is variable."""

    arrayx = ArrayX(name="arrayx", fmt=ascii_zero, dims=(2,))

    def test_nbytes(self):
        with pytest.raises(SizeError):
            return self.arrayx.nbytes



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test integer flag enumeration behavior when value is not a member combination."""

from enum import IntFlag

from baseline import Baseline

from plum.conformance import Case, CaseData
from plum.flag import FlagX

from sample_flag import Register


register = FlagX("register", Register, byteorder="little", nbytes=2)


class TestInvalidValue(Case):

    """Test value is not a member combination."""

    data = CaseData(
        fmt=register,
        bindata=b"\x63\x00",
        nbytes=2,
        values=[Register(99), 99],
        dump=Baseline(
            """
            +--------+--------+-------+-------------------+----------+
            | Offset | Access | Value | Bytes             | Format   |
            +--------+--------+-------+-------------------+----------+
            | 0      |        | 99    | 63 00             | register |
            |  [0]   | .sp    | True  | ........ .......1 | bool     |
            |  [1]   | .r0    | True  | ........ ......1. | bool     |
            |  [2]   | .r1    | False | ........ .....0.. | bool     |
            +--------+--------+-------+-------------------+----------+
            """
        ),
        excess="N/A",
        shortage="N/A",
    )


class Flag(IntFlag):

    """No members."""


flag16 = FlagX("flag16", Flag, byteorder="little", nbytes=2)


class TestNoMember(Case):

    """Test flag type behavior with no members."""

    data = CaseData(
        fmt=flag16,
        bindata=b"\x11\x00",
        nbytes=2,
        values=[Flag(17), 17],
        dump=Baseline(
            """
            +--------+-------+-------+--------+
            | Offset | Value | Bytes | Format |
            +--------+-------+-------+--------+
            | 0      | 17    | 11 00 | flag16 |
            +--------+-------+-------+--------+
            """
        ),
        excess="N/A",
        shortage="N/A",
    )



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Utility functions."""

from typing import Any, List, Tuple

from ._getbytes import getbytes  # pylint: disable=unused-import
from .data import Data, DataMeta
from .dump import Dump
from .exceptions import (
    ExcessMemoryError,
    ImplementationError,
    PackError,
    SizeError,
    UnpackError,
)
from .items import calcsize as _calcsize, items, Format
from .transform import Transform

getbytes.__module__ = "plum"


def calcsize(fmt: Format) -> int:
    """Calculate format size (raise SizeError if variably sized)."""
    if isinstance(fmt, DataMeta):
        nbytes = getattr(fmt, "__nbytes__")

    elif isinstance(fmt, (Data, Transform)):
        nbytes = fmt.nbytes

    else:
        nbytes = _calcsize(fmt)

    if nbytes is None:
        raise SizeError("size varies")

    return nbytes


def pack(value: Any, fmt: Format = None) -> bytes:
    """Pack value as formatted bytes.

    :raises: ``PackError`` if type error, value error, etc.

    """
    pieces: List[bytes] = []
    try:
        # None -> dump
        items.__pack__(value, pieces, None, fmt)
    except Exception as exc:
        # do it over to include dump in exception message
        pack_and_dump(value, fmt)

        raise ImplementationError() from exc  # pragma: no cover

    return b"".join(pieces)


def pack_and_dump(value: Any, fmt: Format = None) -> Tuple[bytes, Dump]:
    """Pack value as formatted bytes and produce bytes summary.

    :raises: ``PackError`` if type error, value error, etc.

    """
    dump = Dump()
    pieces: List[bytes] = []
    try:
        # empty record added so so pack always called with a Record instance
        items.__pack__(value, pieces, dump.add_record(), fmt)

    except Exception as exc:
        dump.trim_blank_record()
        raise PackError(dump=dump, exception=exc) from exc

    dump.trim_blank_record()

    return b"".join(pieces), dump


def unpack(fmt: Format, buffer: bytes) -> Any:
    """Unpack value from formatted bytes.

    :raises: ``UnpackError`` if insufficient bytes or value error

    """
    try:
        # None -> dump
        value, offset = items.__unpack__(buffer, 0, None, fmt)
        if buffer[offset:]:
            raise ExcessMemoryError("too many bytes, retry with dump to follow")
    except Exception as exc:
        # do it over to include dump in exception message
        unpack_and_dump(fmt, buffer)
        raise ImplementationError() from exc  # pragma: no cover

    return value


def unpack_and_dump(fmt: Format, buffer: bytes):
    """Unpack value from formatted bytes and produce a packed bytes summary.

    :raises: ``UnpackError`` if insufficient bytes or value error

    """
    dump = Dump()

    try:
        # empty record added so so pack always called with a Record instance
        dump_record = dump.add_record()

        value, offset = items.__unpack__(buffer, 0, dump_record, fmt)

        extra_bytes = buffer[offset:]

        if extra_bytes:

            for i in range(0, len(extra_bytes), 16):

                record = dump_record.add_record(memory=extra_bytes[i : i + 16])

                if not i:
                    record.separate = True

                    record.value = "<excess bytes>"

            raise ExcessMemoryError(extra_bytes)

    except Exception as exc:
        dump.trim_blank_record()
        raise UnpackError(dump, exc) from exc

    dump.trim_blank_record()

    return value, dump



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Verify view() base method functionality."""

import pytest
from baseline import Baseline

from plum.conformance import wrap_message
from plum.none import NoneX

nil = NoneX("nil")


class TestExceptions:

    """Test view() exceptions."""

    def test_unsupported_plumtype(self):
        """Verify creating a view with an unsupported type raises an exception."""
        with pytest.raises(TypeError) as trap:
            nil.view(buffer=bytearray())

        expected = Baseline("""'nil' does not support view()""")

        assert wrap_message(trap.value) == expected



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test "strict" feature (where strict=True)."""

import pytest

from baseline import Baseline

from plum.conformance import wrap_message
from plum.enum import EnumX
from plum.exceptions import PackError, UnpackError

from sample_enum import Register

register = EnumX("register", Register, 1, "big", signed=False)


class TestPack:

    """Test pack variants with invalid value."""

    expected_message = Baseline(
        """
        +--------+-------+-------+----------+
        | Offset | Value | Bytes | Format   |
        +--------+-------+-------+----------+
        |        | 4     |       | register |
        +--------+-------+-------+----------+

        ValueError occurred during pack operation:

        4 is not a valid Register
        """
    )

    def test_pack(self):
        with pytest.raises(PackError) as trap:
            register.pack(4)

        assert wrap_message(trap.value) == self.expected_message
        assert isinstance(trap.value.__context__, ValueError)

    def test_pack_and_dump(self):
        with pytest.raises(PackError) as trap:
            register.pack_and_dump(4)

        assert wrap_message(trap.value) == self.expected_message
        assert isinstance(trap.value.__context__, ValueError)


class TestUnpack:

    """Test unpack variants with invalid value."""

    expected_message = Baseline(
        """
        +--------+-------+-------+----------+
        | Offset | Value | Bytes | Format   |
        +--------+-------+-------+----------+
        | 0      | 4     | 04    | register |
        +--------+-------+-------+----------+

        ValueError occurred during unpack operation:

        4 is not a valid Register
        """
    )

    def test_unpack(self):
        with pytest.raises(UnpackError) as trap:
            register.unpack(b"\x04")

        assert wrap_message(trap.value) == self.expected_message
        assert isinstance(trap.value.__context__, ValueError)

    def test_unpack_and_dump(self):
        with pytest.raises(UnpackError) as trap:
            register.unpack_and_dump(b"\x04")

        assert wrap_message(trap.value) == self.expected_message
        assert isinstance(trap.value.__context__, ValueError)


class TestTransformCall:

    """Test transform call."""

    def test_call(self):
        expected_message = Baseline(
            """
            4 is not a valid Register
            """
        )
        with pytest.raises(ValueError) as trap:
            register(4)

        assert wrap_message(trap.value) == expected_message



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Integer to bytes and bytes to integer transform."""

from numbers import Integral
from typing import List, Optional, Tuple, Type, Union

from ._getbytes import getbytes
from ._typing import ByteOrderHint
from .data import Data
from .dump import Record
from .exceptions import InsufficientMemoryError
from .transform import Transform
from .view import NumberView


class IntView(NumberView, Integral):  # pylint: disable=too-many-ancestors

    """Integer type view."""

    def __and__(self, other):
        return self.unpack() & other

    def __getitem__(self, item):
        dref = self.__type__.__dref__  # pylint: disable=no-member

        if dref is None:
            raise RuntimeError(
                f"{self.__type__!r} type does not support pointer dereference"
            )

        offset = self.unpack() + (item * dref.__nbytes__)

        return dref.view(self.__buffer__, offset)

    def __iand__(self, other):
        self.set(self.unpack() & other)
        return self

    def __ilshift__(self, other):
        self.set(self.unpack() << other)
        return self

    def __int__(self):
        return NumberView.unpack(self)

    def __invert__(self):
        return ~self.unpack()  # pylint: disable=invalid-unary-operand-type

    def __ior__(self, other):
        self.set(self.unpack() | other)
        return self

    def __irshift__(self, other):
        self.set(self.unpack() >> other)
        return self

    def __ixor__(self, other):
        self.set(self.unpack() ^ other)
        return self

    def __lshift__(self, other):
        return self.unpack() << other

    def __or__(self, other):
        return self.unpack() | other

    def __rand__(self, other):
        return other & self.unpack()

    def __rlshift__(self, other):
        return other << self.unpack()

    def __ror__(self, other):
        return other | self.unpack()

    def __rrshift__(self, other):
        return other >> self.unpack()

    def __rshift__(self, other):
        return self.unpack() >> other

    def __rxor__(self, other):
        return other ^ self.unpack()

    def __setitem__(self, item, value):
        dref_view = self.__getitem__(item)
        dref_view.set(value)

    def __xor__(self, other):
        return self.unpack() ^ other

    def to_bytes(self, length, byteorder, *, signed=False):
        """Return an array of bytes representing an integer.

        :param int length:
            Length of bytes object to use.  An OverflowError is raised if the
            integer is not representable with the given number of bytes.

        :param str byteorder:
            The byte order used to represent the integer.  If byteorder is 'big',
            the most significant byte is at the beginning of the byte array.  If
            byteorder is 'little', the most significant byte is at the end of the
            byte array.  To request the native byte order of the host system, use
            ``sys.byteorder`` as the byte order value.

        :param bool signed:
            Determines whether two's complement is used to represent the integer.
            If signed is False and a negative integer is given, an OverflowError
            is raised.

        :returns: array of bytes
        :rtype: bytes

        """
        # pylint: disable=no-member
        return self.unpack().to_bytes(length, byteorder, signed=signed)


DREF = Optional[Union[Type[Data], Transform]]


class IntX(Transform):

    """Integer to bytes and bytes to integer transform."""

    __dref__: DREF
    __nbytes__: int
    __bns__: Tuple[ByteOrderHint, int, bool]

    def __init__(
        self,
        name: str,
        nbytes: int,
        byteorder: ByteOrderHint = "little",
        *,
        signed: bool = False,
        dref: DREF = None,
    ) -> None:
        super().__init__(name)

        assert nbytes > 0
        assert byteorder in {"big", "little"}

        # store byteorder, nbytes, signed separately for use later
        # to pack/unpack to be able to support any sized integer
        # (versus storing struct pack/unpack function similar to
        # float type which is restricted to common sizes).

        self.__dref__ = dref
        self.__nbytes__ = nbytes
        self.__bns__ = byteorder, nbytes, signed

    @property
    def byteorder(self) -> ByteOrderHint:
        """Byte order ("little" or "big")."""
        return self.__bns__[0]

    @property
    def signed(self) -> bool:
        """Signed integer."""
        return self.__bns__[2]

    def __pack__(
        self, value: int, pieces: List[bytes], dump: Optional[Record] = None
    ) -> None:
        byteorder, nbytes, signed = self.__bns__

        if dump is None:
            pieces.append(value.to_bytes(nbytes, byteorder, signed=signed))
        else:
            # use repr in case str or something that otherwise looks like an int
            dump.value = repr(value)

            try:
                piece = value.to_bytes(nbytes, byteorder, signed=signed)
            except AttributeError:
                raise TypeError(
                    f"value type {type(value)!r} not int-like "
                    f"(no to_bytes() method)"
                ) from None

            dump.value = int(value)
            dump.memory = piece

            pieces.append(piece)

    def __unpack__(
        self, buffer: bytes, offset: int, dump: Optional[Record] = None
    ) -> Tuple[int, int]:
        byteorder, nbytes, signed = self.__bns__

        if dump is None:
            end = offset + nbytes

            if len(buffer) < end:
                raise InsufficientMemoryError("too few bytes to unpack")

            return int.from_bytes(buffer[offset:end], byteorder, signed=signed), end

        chunk, end = getbytes(buffer, offset, dump, nbytes)

        value = dump.value = int.from_bytes(chunk, byteorder, signed=signed)

        return value, end

    def __view__(self, buffer, offset=0):
        """Create integer view of bytes in buffer buffer.

        :param buffer: bytes buffer
        :type buffer: bytes-like (e.g. bytes, bytearray, memoryview)
        :param int offset: byte offset

        """
        return IntView(self, buffer, offset)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test BitFieldMember properties."""

from baseline import Baseline

from plum.structure import Structure, bitfield_member


class Struct(Structure):
    defaulted_bitfield = bitfield_member("doc", size=3)
    kw_bitfield = bitfield_member(
        doc="kw doc",
        size=4,
        lsb=4,
        typ=bool,
        signed=True,
        default=1,
        ignore=True,
        readonly=True,
        compute=False,
        nbytes=2,
        argrepr="argrepr",
    )


class TestDefault:

    """Test with as many left to default as possible."""

    def test_argrepr(self):
        expected = Baseline(
            """
            defaulted_bitfield={self.defaulted_bitfield!r}
            """
        )
        assert Struct.defaulted_bitfield.argrepr == expected

    def test_compute(self):
        assert Struct.defaulted_bitfield.compute is False

    def test_default(self):
        assert Struct.defaulted_bitfield.default is None

    def test_doc(self):
        assert Struct.defaulted_bitfield.doc == "doc"

    def test_ignore(self):
        assert Struct.defaulted_bitfield.ignore is False

    def test_lsb(self):
        assert Struct.defaulted_bitfield.lsb == 0

    def test_name(self):
        assert Struct.defaulted_bitfield.name == "defaulted_bitfield"

    def test_nbytes(self):
        assert Struct.defaulted_bitfield.nbytes == 1

    def test_readonly(self):
        assert Struct.defaulted_bitfield.readonly is False

    def test_signed(self):
        assert Struct.defaulted_bitfield.signed is False

    def test_size(self):
        assert Struct.defaulted_bitfield.size == 3

    def test_typ(self):
        assert Struct.defaulted_bitfield.typ is int


class TestKeyword:

    """Test explicitly defined with keyword argument."""

    def test_argrepr(self):
        assert Struct.kw_bitfield.argrepr == "argrepr"

    def test_compute(self):
        assert Struct.kw_bitfield.compute is False

    def test_default(self):
        assert Struct.kw_bitfield.default == 1

    def test_doc(self):
        assert Struct.kw_bitfield.doc == "kw doc"

    def test_ignore(self):
        assert Struct.kw_bitfield.ignore is True

    def test_lsb(self):
        assert Struct.kw_bitfield.lsb == 4

    def test_name(self):
        assert Struct.kw_bitfield.name == "kw_bitfield"

    def test_nbytes(self):
        assert Struct.kw_bitfield.nbytes == 2

    def test_readonly(self):
        assert Struct.kw_bitfield.readonly is True

    def test_signed(self):
        assert Struct.kw_bitfield.signed is True

    def test_size(self):
        assert Struct.kw_bitfield.size == 4

    def test_typ(self):
        assert Struct.kw_bitfield.typ is bool



"""Test decimal transform properties."""

from plum.decimal import DecimalX


class TestProperties:

    """Test with as many left to default as possible."""

    u16p1 = DecimalX("u16p1", 2, 1, "big", signed=False)

    def test_name(self):
        assert self.u16p1.name == "u16p1"

    def test_nbytes(self):
        assert self.u16p1.nbytes == 2

    def test_byteorder(self):
        assert self.u16p1.byteorder == "big"

    def test_signed(self):
        assert self.u16p1.signed is False

    def test_precision(self):
        assert self.u16p1.precision == 1



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Structure __init__ method generator."""

import atexit
import inspect

from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union

from .dump import Dump, Record


class CodeInjector:

    """Code evaluator/injector."""

    def __init__(self, namespace: Dict[str, Any]) -> None:
        self.f_globals = inspect.stack()[2].frame.f_globals
        self.namespace = namespace
        self.annotations = namespace.get("__annotations__", {})

    def evaluate(self, expression: str):
        """Evaluate expression within namespaces of class."""
        # pylint: disable=eval-used
        try:
            result = eval(expression, self.f_globals)
        except TypeError:
            # typeerror: e.g. Optional[int]
            result = object()

        return result

    def get_expression(self, value: Any, alternative: str) -> str:
        """Get expression which will evaluate to be the identity of a value.

        :param value: Python object to generate expression for
        :param alternative: express to use if it does not evaluate properly

        """
        # FUTURE - either eliminate evaluate() or get_expression()
        if isinstance(value, int) and not isinstance(value, Enum):
            expression = str(int(value))
            is_same = True

        else:
            if isinstance(value, type):
                expression = value.__name__
            elif isinstance(value, Enum):
                expression = str(value)
            else:
                expression = repr(value)

            try:
                is_same = self.evaluate(expression) is value
            except (NameError, SyntaxError):
                is_same = False

        return expression if is_same else alternative

    def execute_lines(self, lines: List[str]) -> None:
        """Execute lines within the same namespaces of a class."""
        f_globals = self.f_globals

        extra_globals = {
            "Any": Any,
            "Dump": Dump,
            "List": List,
            "Optional": Optional,
            "Record": Record,
            "Tuple": Tuple,
            "Union": Union,
        }

        for key in list(extra_globals):
            if key in f_globals:
                del extra_globals[key]
        f_globals.update(extra_globals)

        try:
            # pylint: disable=exec-used
            exec("\n".join(lines), f_globals, self.namespace)
        except Exception:  # pragma: no cover
            for i, line in enumerate(lines):
                print(f"{i + 1:04d}: {line}")
            raise

        for key in extra_globals:
            del f_globals[key]

    def get_type_hint_expression(self, name: str) -> str:  # pragma: no cover
        """Get expression which will evaluate to the type annotation for a member.

        :param name: member name

        """
        try:
            python_type = self.annotations[name]
        except KeyError:
            return ""

        if isinstance(python_type, str):
            return python_type

        # sys.version_info >= (3, 10) never gets beyond this point (hence no cover)

        try:
            type_hint = python_type.__name__
        except AttributeError:

            def fixup(text: str, separators=("[", "]")):
                if separators:
                    sep = separators[0]
                    text = sep.join(fixup(x, separators[1:]) for x in text.split(sep))
                return text

            type_hint = fixup(
                str(python_type)
            )  # e.g. "typing.Any", "List[xyz.Segment]"

        try:
            # pylint: disable=eval-used
            is_same = eval(type_hint, self.f_globals, self.namespace) is python_type
        except NameError:  # pragma: no cover
            return ""
        except Exception as exc:
            raise RuntimeError(f"invalid type hint: {type_hint!r}") from exc

        return type_hint if is_same else ""

    updates: Dict[str, List[Tuple[int, List[str], List[str]]]] = {}

    @staticmethod
    def get_lines(
        indent: str, lines: List[str], selections: List[str]
    ):  # pragma: no cover
        """Get selected methods implementation lines."""
        indented_lines = []
        if selections:
            decorator: Optional[str] = None

            keep_lines = False

            for line in lines:
                line = line.rstrip()
                if keep_lines and (not line or line.startswith(" ")):
                    indented_lines.append(indent + line)
                    continue

                keep_lines = False

                if line.startswith("@"):
                    decorator = line
                    continue

                if line.startswith(" ") or not line:
                    continue

                if line.startswith("def "):

                    function_name = line[4:].replace("(", " ").split()[0].strip("_")

                    if function_name in selections:
                        keep_lines = True
                        if decorator:
                            indented_lines.append(indent + decorator)

                        indented_lines.append(indent + line)

                    decorator = None
                    continue

                if line.startswith("__"):  # e.g. __eq__ = list.__eq__
                    function_name = line.split()[0].strip("_")
                    if function_name in selections:
                        keep_lines = True
                        indented_lines.append(indent + line)

        else:
            indented_lines += [indent + line for line in lines]

        return [line.rstrip() for line in indented_lines]

    @classmethod
    def update_scripts(cls):
        """Insert method implementation code into scripts where designated.

        Insert code into class definitions where "Structure.implementation" found.
        """
        for path, update in cls.updates.items():  # pragma: no cover
            with open(path) as script:
                script_lines = script.read().split("\n")

            for lineno, lines, selections in reversed(update):
                line = script_lines[lineno - 1]
                indent = line[: len(line) - len(line.lstrip())]
                script_lines[lineno - 1 : lineno] = cls.get_lines(
                    indent, lines, selections
                )

            with open(path, "w") as script:
                script.write("\n".join(script_lines))

    def update_script(self, lines):
        """Insert method implementation code into script at designated spot.

        Insert code into class definition where "Structure.implementation" found.

        """
        try:
            # e.g. "Structure.implementation" in the class body puts this in namespace
            path, lineno, selections = self.namespace["__implementation__"]

        except KeyError:
            pass

        else:  # pragma: no cover
            if not self.updates:
                atexit.register(self.update_scripts)

            self.updates.setdefault(path, []).append((lineno, lines, selections))



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test dimmed_member() support of cls factory."""

import sys
from enum import IntEnum

import pytest

from baseline import Baseline

from plum.array import ArrayX
from plum.conformance import Case, CaseData, wrap_message
from plum.enum import EnumX
from plum.littleendian import uint8, uint16
from plum.structure import Structure, dimmed_member, member

uint8array = ArrayX("uint8array", fmt=uint8)
uint16array = ArrayX("uint16array", fmt=uint16)


class DataType(IntEnum):

    """Array type."""

    UINT8 = 0
    UINT16 = 1


datatype = EnumX("datatype", DataType, nbytes=1, byteorder="big", signed=False)


array_map = {
    DataType.UINT8: uint8array,
    DataType.UINT16: uint16array,
}


def _get_array_cls(parent):
    return array_map[parent[0]]


def _get_data_type(parent):
    if max(parent[2]) < 256:
        cls = DataType.UINT8
    else:
        cls = DataType.UINT16
    return cls


class TestSimpleSingleDim(Case):

    """Test dimmed member has single dim."""

    class SampleStruct(Structure):

        data_len: int = member(fmt=uint8)
        data: list = dimmed_member(dims=data_len, fmt=uint8array)
        bookend: int = member(fmt=uint8)

    data = CaseData(
        fmt=SampleStruct,
        bindata=bytes.fromhex("040001020377"),
        nbytes=None,
        values=(
            SampleStruct(data_len=4, data=[0, 1, 2, 3], bookend=0x77),
            [4, [0, 1, 2, 3], 0x77],
        ),
        dump=Baseline(
            """
            +--------+----------+-------+-------+--------------+
            | Offset | Access   | Value | Bytes | Format       |
            +--------+----------+-------+-------+--------------+
            |        |          |       |       | SampleStruct |
            | 0      | data_len | 4     | 04    | uint8        |
            |        | data     |       |       | uint8array   |
            | 1      |   [0]    | 0     | 00    | uint8        |
            | 2      |   [1]    | 1     | 01    | uint8        |
            | 3      |   [2]    | 2     | 02    | uint8        |
            | 4      |   [3]    | 3     | 03    | uint8        |
            | 5      | bookend  | 119   | 77    | uint8        |
            +--------+----------+-------+-------+--------------+
            """
        ),
        excess=Baseline(
            """
            +--------+----------+----------------+-------+--------------+
            | Offset | Access   | Value          | Bytes | Format       |
            +--------+----------+----------------+-------+--------------+
            |        |          |                |       | SampleStruct |
            | 0      | data_len | 4              | 04    | uint8        |
            |        | data     |                |       | uint8array   |
            | 1      |   [0]    | 0              | 00    | uint8        |
            | 2      |   [1]    | 1              | 01    | uint8        |
            | 3      |   [2]    | 2              | 02    | uint8        |
            | 4      |   [3]    | 3              | 03    | uint8        |
            | 5      | bookend  | 119            | 77    | uint8        |
            +--------+----------+----------------+-------+--------------+
            | 6      |          | <excess bytes> | 99    |              |
            +--------+----------+----------------+-------+--------------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
            +--------+----------+----------------------+-------+--------------+
            | Offset | Access   | Value                | Bytes | Format       |
            +--------+----------+----------------------+-------+--------------+
            |        |          |                      |       | SampleStruct |
            | 0      | data_len | 4                    | 04    | uint8        |
            |        | data     |                      |       | uint8array   |
            | 1      |   [0]    | 0                    | 00    | uint8        |
            | 2      |   [1]    | 1                    | 01    | uint8        |
            | 3      |   [2]    | 2                    | 02    | uint8        |
            | 4      |   [3]    | 3                    | 03    | uint8        |
            |        | bookend  | <insufficient bytes> |       | uint8        |
            +--------+----------+----------------------+-------+--------------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack 'uint8' (1 needed, only 0 available)
            """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, data_len: int, data: list, bookend: int) -> None:
                self[:] = (data_len, data, bookend)

            @data_len.getter
            def data_len(self) -> int:
                return self[0]

            @data_len.setter
            def data_len(self, value: int) -> None:
                self[0] = value

            @data.getter
            def data(self) -> list:
                return self[1]

            @data.setter
            def data(self, value: list) -> None:
                self[1] = value

            @bookend.getter
            def bookend(self) -> int:
                return self[2]

            @bookend.setter
            def bookend(self, value: int) -> None:
                self[2] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_data_len, m_data, m_bookend) = value

                if dump is None:
                    uint8.__pack__(m_data_len, pieces, dump)

                    uint8array.__pack__(m_data, pieces, dump, dims=(m_data_len, ))

                    uint8.__pack__(m_bookend, pieces, dump)

                else:
                    data_len_dump = dump.add_record(access="data_len", fmt=uint8)
                    uint8.__pack__(m_data_len, pieces, data_len_dump)

                    data_dump = dump.add_record(access="data", fmt=uint8array)
                    uint8array.__pack__(m_data, pieces, data_dump, dims=(m_data_len, ))

                    bookend_dump = dump.add_record(access="bookend", fmt=uint8)
                    uint8.__pack__(m_bookend, pieces, bookend_dump)

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["SampleStruct", int]:
                structure = list.__new__(cls)

                if dump is None:
                    m_data_len, offset = uint8.__unpack__(buffer, offset, dump)

                    m_data, offset = uint8array.__unpack__(buffer, offset, dump, dims=(m_data_len, ))

                    m_bookend, offset = uint8.__unpack__(buffer, offset, dump)

                else:
                    data_len_dump = dump.add_record(access="data_len", fmt=uint8)
                    m_data_len, offset = uint8.__unpack__(buffer, offset, data_len_dump)

                    data_dump = dump.add_record(access="data", fmt=uint8array)
                    m_data, offset = uint8array.__unpack__(buffer, offset, data_dump, dims=(m_data_len, ))

                    bookend_dump = dump.add_record(access="bookend", fmt=uint8)
                    m_bookend, offset = uint8.__unpack__(buffer, offset, bookend_dump)

                structure[:] = (m_data_len, m_data, m_bookend)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(data_len={self.data_len!r}, data={self.data!r}, bookend={self.bookend!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )


class TestTypeFactorySingleDim(Case):

    """Test dimmed member has single dim and variable type."""

    class SampleStruct(Structure):

        data_type: int = member(fmt=datatype, default=_get_data_type)
        data_len: int = member(fmt=uint8, compute=True)
        data: list = dimmed_member(dims=data_len, fmt=_get_array_cls)
        bookend: int = member(fmt=uint8)

    data = CaseData(
        fmt=SampleStruct,
        bindata=bytes.fromhex("00040001020377"),
        nbytes=None,
        values=(
            SampleStruct(
                data_type=DataType.UINT8, data_len=4, data=[0, 1, 2, 3], bookend=0x77
            ),
            SampleStruct(data=[0, 1, 2, 3], bookend=0x77),
            [DataType.UINT8, 4, [0, 1, 2, 3], 0x77],
        ),
        dump=Baseline(
            """
            +--------+-----------+----------------+-------+--------------+
            | Offset | Access    | Value          | Bytes | Format       |
            +--------+-----------+----------------+-------+--------------+
            |        |           |                |       | SampleStruct |
            | 0      | data_type | DataType.UINT8 | 00    | datatype     |
            | 1      | data_len  | 4              | 04    | uint8        |
            |        | data      |                |       | uint8array   |
            | 2      |   [0]     | 0              | 00    | uint8        |
            | 3      |   [1]     | 1              | 01    | uint8        |
            | 4      |   [2]     | 2              | 02    | uint8        |
            | 5      |   [3]     | 3              | 03    | uint8        |
            | 6      | bookend   | 119            | 77    | uint8        |
            +--------+-----------+----------------+-------+--------------+
            """
        ),
        excess=Baseline(
            """
            +--------+-----------+----------------+-------+--------------+
            | Offset | Access    | Value          | Bytes | Format       |
            +--------+-----------+----------------+-------+--------------+
            |        |           |                |       | SampleStruct |
            | 0      | data_type | DataType.UINT8 | 00    | datatype     |
            | 1      | data_len  | 4              | 04    | uint8        |
            |        | data      |                |       | uint8array   |
            | 2      |   [0]     | 0              | 00    | uint8        |
            | 3      |   [1]     | 1              | 01    | uint8        |
            | 4      |   [2]     | 2              | 02    | uint8        |
            | 5      |   [3]     | 3              | 03    | uint8        |
            | 6      | bookend   | 119            | 77    | uint8        |
            +--------+-----------+----------------+-------+--------------+
            | 7      |           | <excess bytes> | 99    |              |
            +--------+-----------+----------------+-------+--------------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
            +--------+-----------+----------------------+-------+--------------+
            | Offset | Access    | Value                | Bytes | Format       |
            +--------+-----------+----------------------+-------+--------------+
            |        |           |                      |       | SampleStruct |
            | 0      | data_type | DataType.UINT8       | 00    | datatype     |
            | 1      | data_len  | 4                    | 04    | uint8        |
            |        | data      |                      |       | uint8array   |
            | 2      |   [0]     | 0                    | 00    | uint8        |
            | 3      |   [1]     | 1                    | 01    | uint8        |
            | 4      |   [2]     | 2                    | 02    | uint8        |
            | 5      |   [3]     | 3                    | 03    | uint8        |
            |        | bookend   | <insufficient bytes> |       | uint8        |
            +--------+-----------+----------------------+-------+--------------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack 'uint8' (1 needed, only 0 available)
            """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, data_type: Optional[int] = None, data_len: Optional[int] = None, data: list, bookend: int) -> None:
                if data_len is None:
                    data_len = len(data)

                self[:] = (data_type, data_len, data, bookend)

                if data_type is None:
                    self[0] = type(self).data_type.default(self)

            @data_type.getter
            def data_type(self) -> int:
                return self[0]

            @data_type.setter
            def data_type(self, value: int) -> None:
                self[0] = value

            @data_len.getter
            def data_len(self) -> int:
                return self[1]

            @data_len.setter
            def data_len(self, value: int) -> None:
                self[1] = value

            @data.getter
            def data(self) -> list:
                return self[2]

            @data.setter
            def data(self, value: list) -> None:
                self[1] = len(value)  # update 'data_len' member
                self[2] = value

            @bookend.getter
            def bookend(self) -> int:
                return self[3]

            @bookend.setter
            def bookend(self, value: int) -> None:
                self[3] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_data_type, m_data_len, m_data, m_bookend) = value

                if dump is None:
                    datatype.__pack__(m_data_type, pieces, dump)

                    uint8.__pack__(m_data_len, pieces, dump)

                    data_fmt = cls.data.fmt(value)
                    data_fmt.__pack__(m_data, pieces, dump, dims=(m_data_len, ))

                    uint8.__pack__(m_bookend, pieces, dump)

                else:
                    data_type_dump = dump.add_record(access="data_type", fmt=datatype)
                    datatype.__pack__(m_data_type, pieces, data_type_dump)

                    data_len_dump = dump.add_record(access="data_len", fmt=uint8)
                    uint8.__pack__(m_data_len, pieces, data_len_dump)

                    data_fmt = cls.data.fmt(value)
                    data_dump = dump.add_record(access="data", fmt=data_fmt)
                    data_fmt.__pack__(m_data, pieces, data_dump, dims=(m_data_len, ))

                    bookend_dump = dump.add_record(access="bookend", fmt=uint8)
                    uint8.__pack__(m_bookend, pieces, bookend_dump)

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["SampleStruct", int]:
                structure = list.__new__(cls)
                add_member = structure.append

                if dump is None:
                    m_data_type, offset = datatype.__unpack__(buffer, offset, dump)
                    add_member(m_data_type)

                    m_data_len, offset = uint8.__unpack__(buffer, offset, dump)
                    add_member(m_data_len)

                    data_fmt = cls.data.fmt(structure)
                    m_data, offset = data_fmt.__unpack__(buffer, offset, dump, dims=(m_data_len, ))
                    add_member(m_data)

                    m_bookend, offset = uint8.__unpack__(buffer, offset, dump)
                    add_member(m_bookend)

                else:
                    data_type_dump = dump.add_record(access="data_type", fmt=datatype)
                    m_data_type, offset = datatype.__unpack__(buffer, offset, data_type_dump)
                    add_member(m_data_type)

                    data_len_dump = dump.add_record(access="data_len", fmt=uint8)
                    m_data_len, offset = uint8.__unpack__(buffer, offset, data_len_dump)
                    add_member(m_data_len)

                    data_fmt = cls.data.fmt(structure)
                    data_dump = dump.add_record(access="data", fmt=data_fmt)
                    m_data, offset = data_fmt.__unpack__(buffer, offset, data_dump, dims=(m_data_len, ))
                    add_member(m_data)

                    bookend_dump = dump.add_record(access="bookend", fmt=uint8)
                    m_bookend, offset = uint8.__unpack__(buffer, offset, bookend_dump)
                    add_member(m_bookend)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                elif isinstance(other, type(self)):
                    _s_data_type, _s_data_len, _s_data, _s_bookend = self
                    if _s_data_len is None:
                        _s_data_type, _s_data_len, _s_data, _s_bookend = self.unpack(self.ipack())
                    _o_data_type, _o_data_len, _o_data, _o_bookend = other
                    if _o_data_len is None:
                        _o_data_type, _o_data_len, _o_data, _o_bookend = self.unpack(other.ipack())
                    return (_s_data_type, _s_data_len, _s_data, _s_bookend) == (_o_data_type, _o_data_len, _o_data, _o_bookend)
                else:
                    return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                elif isinstance(other, type(self)):
                    _s_data_type, _s_data_len, _s_data, _s_bookend = self
                    if _s_data_len is None:
                        _s_data_type, _s_data_len, _s_data, _s_bookend = self.unpack(self.ipack())
                    _o_data_type, _o_data_len, _o_data, _o_bookend = other
                    if _o_data_len is None:
                        _o_data_type, _o_data_len, _o_data, _o_bookend = self.unpack(other.ipack())
                    return (_s_data_type, _s_data_len, _s_data, _s_bookend) != (_o_data_type, _o_data_len, _o_data, _o_bookend)
                else:
                    return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(data_type={str(self.data_type)}, data_len={self.data_len!r}, data={self.data!r}, bookend={self.bookend!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )


class TestTypeFactoryMultiDim(Case):

    """Test dimmed member has multipe dims and variable type."""

    class MultiDimStruct(Structure):

        data_type: int = member(fmt=datatype, default=lambda parent: DataType.UINT8)
        data_dims: list = member(fmt=ArrayX("dims", fmt=uint8, dims=(2,)), compute=True)
        data: list = dimmed_member(dims=data_dims, fmt=_get_array_cls)
        bookend: int = member(fmt=uint8)

    data = CaseData(
        fmt=MultiDimStruct,
        bindata=bytes.fromhex("0002020001020377"),
        nbytes=None,
        values=(
            MultiDimStruct(
                data_type=DataType.UINT8,
                data_dims=[2, 2],
                data=[[0, 1], [2, 3]],
                bookend=0x77,
            ),
            MultiDimStruct(data=[[0, 1], [2, 3]], bookend=0x77),
            [DataType.UINT8, [2, 2], [[0, 1], [2, 3]], 0x77],
        ),
        dump=Baseline(
            """
            +--------+-----------+----------------+-------+----------------+
            | Offset | Access    | Value          | Bytes | Format         |
            +--------+-----------+----------------+-------+----------------+
            |        |           |                |       | MultiDimStruct |
            | 0      | data_type | DataType.UINT8 | 00    | datatype       |
            |        | data_dims |                |       | dims           |
            | 1      |   [0]     | 2              | 02    | uint8          |
            | 2      |   [1]     | 2              | 02    | uint8          |
            |        | data      |                |       | uint8array     |
            |        |   [0]     |                |       |                |
            | 3      |     [0]   | 0              | 00    | uint8          |
            | 4      |     [1]   | 1              | 01    | uint8          |
            |        |   [1]     |                |       |                |
            | 5      |     [0]   | 2              | 02    | uint8          |
            | 6      |     [1]   | 3              | 03    | uint8          |
            | 7      | bookend   | 119            | 77    | uint8          |
            +--------+-----------+----------------+-------+----------------+
            """
        ),
        excess=Baseline(
            """
            +--------+-----------+----------------+-------+----------------+
            | Offset | Access    | Value          | Bytes | Format         |
            +--------+-----------+----------------+-------+----------------+
            |        |           |                |       | MultiDimStruct |
            | 0      | data_type | DataType.UINT8 | 00    | datatype       |
            |        | data_dims |                |       | dims           |
            | 1      |   [0]     | 2              | 02    | uint8          |
            | 2      |   [1]     | 2              | 02    | uint8          |
            |        | data      |                |       | uint8array     |
            |        |   [0]     |                |       |                |
            | 3      |     [0]   | 0              | 00    | uint8          |
            | 4      |     [1]   | 1              | 01    | uint8          |
            |        |   [1]     |                |       |                |
            | 5      |     [0]   | 2              | 02    | uint8          |
            | 6      |     [1]   | 3              | 03    | uint8          |
            | 7      | bookend   | 119            | 77    | uint8          |
            +--------+-----------+----------------+-------+----------------+
            | 8      |           | <excess bytes> | 99    |                |
            +--------+-----------+----------------+-------+----------------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
            +--------+-----------+----------------------+-------+----------------+
            | Offset | Access    | Value                | Bytes | Format         |
            +--------+-----------+----------------------+-------+----------------+
            |        |           |                      |       | MultiDimStruct |
            | 0      | data_type | DataType.UINT8       | 00    | datatype       |
            |        | data_dims |                      |       | dims           |
            | 1      |   [0]     | 2                    | 02    | uint8          |
            | 2      |   [1]     | 2                    | 02    | uint8          |
            |        | data      |                      |       | uint8array     |
            |        |   [0]     |                      |       |                |
            | 3      |     [0]   | 0                    | 00    | uint8          |
            | 4      |     [1]   | 1                    | 01    | uint8          |
            |        |   [1]     |                      |       |                |
            | 5      |     [0]   | 2                    | 02    | uint8          |
            | 6      |     [1]   | 3                    | 03    | uint8          |
            |        | bookend   | <insufficient bytes> |       | uint8          |
            +--------+-----------+----------------------+-------+----------------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack 'uint8' (1 needed, only 0 available)
            """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, data_type: Optional[int] = None, data_dims: Optional[list] = None, data: list, bookend: int) -> None:
                if data_dims is None:
                    data_dims = list(type(self).data.compute_dims(data, 2))

                self[:] = (data_type, data_dims, data, bookend)

                if data_type is None:
                    self[0] = type(self).data_type.default(self)

            @data_type.getter
            def data_type(self) -> int:
                return self[0]

            @data_type.setter
            def data_type(self, value: int) -> None:
                self[0] = value

            @data_dims.getter
            def data_dims(self) -> list:
                return self[1]

            @data_dims.setter
            def data_dims(self, value: list) -> None:
                self[1] = value

            @data.getter
            def data(self) -> list:
                return self[2]

            @data.setter
            def data(self, value: list) -> None:
                self[1] = list(type(self).data.compute_dims(value, 2))  # update 'data_dims' member
                self[2] = value

            @bookend.getter
            def bookend(self) -> int:
                return self[3]

            @bookend.setter
            def bookend(self, value: int) -> None:
                self[3] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_data_type, m_data_dims, m_data, m_bookend) = value

                if dump is None:
                    datatype.__pack__(m_data_type, pieces, dump)

                    cls.data_dims.fmt.__pack__(m_data_dims, pieces, dump)

                    data_fmt = cls.data.fmt(value)
                    data_fmt.__pack__(m_data, pieces, dump, dims=m_data_dims)

                    uint8.__pack__(m_bookend, pieces, dump)

                else:
                    data_type_dump = dump.add_record(access="data_type", fmt=datatype)
                    datatype.__pack__(m_data_type, pieces, data_type_dump)

                    data_dims_dump = dump.add_record(access="data_dims", fmt=cls.data_dims.fmt)
                    cls.data_dims.fmt.__pack__(m_data_dims, pieces, data_dims_dump)

                    data_fmt = cls.data.fmt(value)
                    data_dump = dump.add_record(access="data", fmt=data_fmt)
                    data_fmt.__pack__(m_data, pieces, data_dump, dims=m_data_dims)

                    bookend_dump = dump.add_record(access="bookend", fmt=uint8)
                    uint8.__pack__(m_bookend, pieces, bookend_dump)

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["MultiDimStruct", int]:
                structure = list.__new__(cls)
                add_member = structure.append

                if dump is None:
                    m_data_type, offset = datatype.__unpack__(buffer, offset, dump)
                    add_member(m_data_type)

                    m_data_dims, offset = cls.data_dims.fmt.__unpack__(buffer, offset, dump)
                    add_member(m_data_dims)

                    data_fmt = cls.data.fmt(structure)
                    m_data, offset = data_fmt.__unpack__(buffer, offset, dump, dims=m_data_dims)
                    add_member(m_data)

                    m_bookend, offset = uint8.__unpack__(buffer, offset, dump)
                    add_member(m_bookend)

                else:
                    data_type_dump = dump.add_record(access="data_type", fmt=datatype)
                    m_data_type, offset = datatype.__unpack__(buffer, offset, data_type_dump)
                    add_member(m_data_type)

                    data_dims_dump = dump.add_record(access="data_dims", fmt=cls.data_dims.fmt)
                    m_data_dims, offset = cls.data_dims.fmt.__unpack__(buffer, offset, data_dims_dump)
                    add_member(m_data_dims)

                    data_fmt = cls.data.fmt(structure)
                    data_dump = dump.add_record(access="data", fmt=data_fmt)
                    m_data, offset = data_fmt.__unpack__(buffer, offset, data_dump, dims=m_data_dims)
                    add_member(m_data)

                    bookend_dump = dump.add_record(access="bookend", fmt=uint8)
                    m_bookend, offset = uint8.__unpack__(buffer, offset, bookend_dump)
                    add_member(m_bookend)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                elif isinstance(other, type(self)):
                    _s_data_type, _s_data_dims, _s_data, _s_bookend = self
                    if _s_data_dims is None:
                        _s_data_type, _s_data_dims, _s_data, _s_bookend = self.unpack(self.ipack())
                    _o_data_type, _o_data_dims, _o_data, _o_bookend = other
                    if _o_data_dims is None:
                        _o_data_type, _o_data_dims, _o_data, _o_bookend = self.unpack(other.ipack())
                    return (_s_data_type, _s_data_dims, _s_data, _s_bookend) == (_o_data_type, _o_data_dims, _o_data, _o_bookend)
                else:
                    return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                elif isinstance(other, type(self)):
                    _s_data_type, _s_data_dims, _s_data, _s_bookend = self
                    if _s_data_dims is None:
                        _s_data_type, _s_data_dims, _s_data, _s_bookend = self.unpack(self.ipack())
                    _o_data_type, _o_data_dims, _o_data, _o_bookend = other
                    if _o_data_dims is None:
                        _o_data_type, _o_data_dims, _o_data, _o_bookend = self.unpack(other.ipack())
                    return (_s_data_type, _s_data_dims, _s_data, _s_bookend) != (_o_data_type, _o_data_dims, _o_data, _o_bookend)
                else:
                    return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(data_type={str(self.data_type)}, data_dims={self.data_dims!r}, data={self.data!r}, bookend={self.bookend!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )

    expected_dump = Baseline(
        """
        +--------+---------------+----------------+-------+----------------+
        | Offset | Access        | Value          | Bytes | Format         |
        +--------+---------------+----------------+-------+----------------+
        |        |               |                |       | MultiDimStruct |
        | 0      | data_type [0] | DataType.UINT8 | 00    | datatype       |
        |        | data_dims [1] |                |       | dims           |
        | 1      |   [0]         | 2              | 02    | uint8          |
        | 2      |   [1]         | 2              | 02    | uint8          |
        |        | data [2]      |                |       | uint8array     |
        |        |   [0]         |                |       |                |
        | 3      |     [0]       | 0              | 00    | uint8          |
        | 4      |     [1]       | 1              | 01    | uint8          |
        |        |   [1]         |                |       |                |
        | 5      |     [0]       | 2              | 02    | uint8          |
        | 6      |     [1]       | 3              | 03    | uint8          |
        +--------+---------------+----------------+-------+----------------+
        """
    )


class TestExceptions:

    """Test exception use cases."""

    def test_bad_dims_member(self):
        """Verify exception/message if dimmed_member() gets bad dims member."""
        with pytest.raises(TypeError) as trap:

            class BadRef(Structure):  # pylint: disable=unused-variable

                """Simple sized array with bad reference."""

                count: int = member(fmt=uint8, compute=True)
                array: list = dimmed_member(fmt=ArrayX("array", fmt=uint8), dims="junk")
                bookend: uint8

        expected = Baseline(
            """
            invalid 'dims', must be a 'member()'
            """
        )

        assert wrap_message(trap.value) == expected

    def test_missing_array_member(self):
        """Verify exception/message if missing dimmed_member() definition."""
        with pytest.raises(TypeError) as trap:

            class BadRef(Structure):  # pylint: disable=unused-variable

                """Simple sized array with missing array member."""

                count: int = member(fmt=uint8, compute=True)

        expected = Baseline(
            """
            'count' member never associated with member used to compute it
            """
        )

        assert wrap_message(trap.value) == expected


class TestException:
    def test_dims_wrong_type(self):
        expected_message = Baseline(
            """
            'dims' dims member not an integer or array transform format
            """
        )

        with pytest.raises(TypeError) as trap:

            class Struct(Structure):  # pylint: disable=unused-variable
                dims = member(fmt=datatype, compute=True)
                array = dimmed_member(fmt=uint8array, dims=dims)

        assert str(trap.value) == expected_message


class TestSetter:
    def test_override(self):
        class Struct(Structure):
            dims = member(fmt=uint8, compute=True)
            array = dimmed_member(fmt=uint8array, dims=dims)

            @array.setter
            def array(self, value):
                # pylint: disable=unused-argument
                self[1] = [99]

        struct = Struct(dims=1, array=[0])
        struct.array = [1, 2, 3]
        assert struct.array == [99]

    def test_readonly(self):
        expected_message = Baseline(
            """
            can't set attribute 'array'
            """
        )

        class Struct(Structure):
            dims = member(fmt=uint8, compute=True)
            array = dimmed_member(fmt=uint8array, dims=dims, readonly=True)

        struct = Struct(dims=1, array=[0])

        with pytest.raises(AttributeError) as trap:
            struct.array = [1, 2, 3]

        actual_message = str(trap.value)

        if sys.version_info < (3, 10):
            actual_message += " 'array'"

        assert actual_message == expected_message
